'''index_fasta.py - Index fasta formatted files
============================================

:Tags: Genomics Sequences FASTA Manipulation

Purpose
-------

This script indexes one or more :term:`fasta` formatted files into a
database that can be used by other scripts in the cgat code collection
and :mod:`IndexedFasta` for quick access to a particular part of a sequence.
This is very useful for large genomic sequences.

By default, the database is itself a :term:`fasta` formatted file in
which all line breaks and other white space characters have been
removed.  Compression methods are available to conserve disk space,
though they do come at a performance penalty.

The script implements several indexing and compression methods.  The
default method uses no compression and builds a simple random access
index based on a table of absolute file positions.  The sequence is
stored in a plain fasta file with one line per sequence allowing to
extract a sequence segment by direct file positioning.

Alternatively, the sequence can be block-compressed using different
compression methods (gzip, lzo, bzip). These are mostly for research
purposes.

See also http://pypi.python.org/pypi/pyfasta for another
implementation.  Samtools provides similar functionality with the
``samtools faidx`` command and block compression has been implemented
in the `bgzip http://samtools.sourceforge.net/tabix.shtml>`_ tool.

The script permits supplying synonyms to the database index. For
example, setting ``--synonyms=chrM=chrMT`` will ensure that the
mitochondrial genome sequence is returned both for the keys ``chrM``
and ``chrMT``.

Examples
--------

Index a collection of fasta files in a compressed archive::

   python index_fasta.py oa_ornAna1_softmasked ornAna1.fa.gz > oa_ornAna1_softmasked.log

To retrieve a segment::

   python index_fasta.py --extract=chr5:1000:2000 oa_ornAna1_softmasked

Indexing from a tar file is possible::

   python index_fasta.py oa_ornAna1_softmasked ornAna1.tar.gz > oa_ornAna1_softmasked.log

Indexing from stdin requires to use the ``-`` place-holder::

   zcat ornAna1.fa.gz | python index_fasta.py oa_ornAna1_softmasked - > oa_ornAna1_softmasked.log

Usage
-----

Type::

   cgat index_genome DATABASE [SOURCE...|-] [OPTIONS]
   cgat index_genome DATABASE [SOURCE...|-] --compression=COMPRESSION --random-access-points=100000

To create indexed DATABASE from SOURCE. Supply - as SOURCE to read from stdin.
If the output is to be compressed, a spacing for the random access points must
be supplied.

Type::

   cgat index_genome DATABASE --extract=CONTIG:[STRAND]:START:END

To extract the bases on the STRAND strand, between START to END from
entry CONTIG, from DATABASE.

Command line options
--------------------

'''
import cgat.IndexedFasta as IndexedFasta
import cgatcore.experiment as E
import sys


def main(argv=None):

    if argv is None:
        argv = sys.argv

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument(
        "-e", "--extract", dest="extract", type=str,
        help="extract region for testing purposes. Format is "
        "contig:strand:from:to. "
        "The default coordinates are 0-based "
        "open/closed coordinates on both strands, but can be changed "
        "by --input-format. "
        "For example, 'chr1:+:10:12' will return "
        "bases 11 and 12 on chr1. Elements from the end of the "
        "string can be omitted. For example, 'chr1' will return "
        "all of chromosome 'chr1'.")

    input_format_choices = ("one-forward-open", "zero-both-open")
    parser.add_argument("-i", "--input-format", dest="input_format",
                        type=str,
                        choices=input_format_choices,
                        help="coordinate format of input. Valid choices are "
                        "%s. See --extract." %
                        ", ".join(input_format_choices))

    parser.add_argument(
        "-s", "--synonyms", dest="synonyms", type=str,
        help="list of synonyms. This is a comma separated with list "
        "of equivalence relations. For example, chrM=chrMT "
        "means that chrMT will refer to chrM and either "
        "can be used to retrieve a sequence ")

    group = parser.add_argument_group("Bencharking options")

    group.add_argument("-b", "--benchmark", dest="benchmark",
                       action="store_true",
                       help="benchmark time for read access ")
    group.add_argument("--benchmark-num-iterations",
                       dest="benchmark_num_iterations",
                       type=int,
                       help="number of iterations for benchmark ")
    group.add_argument("--benchmark-fragment-size",
                       dest="benchmark_fragment_size",
                       type=int,
                       help="benchmark: fragment size.")
    parser.add_argument_group(group)

    group = parser.add_argument_group("Validation options")

    group.add_argument("--verify", dest="verify", type=str,
                       help="verify against other database.")

    group.add_argument("--verify-iterations", dest="verify_num_iterations",
                       type=int,
                       help="number of iterations for verification ")
    parser.add_argument_group(group)

    file_format_choices = ("fasta", "auto", "fasta.gz", "tar", "tar.gz")
    parser.add_argument("--file-format", dest="file_format", type=str,
                        choices=file_format_choices,
                        help="file format of input. Supply if data comes "
                        "from stdin "
                        "Valid choices are \\%s." %
                        ", ".join(file_format_choices))

    parser.add_argument("-a", "--clean-sequence", dest="clean_sequence",
                        action="store_true",
                        help="remove X/x from DNA sequences - they cause "
                        "errors in exonerate.")

    parser.add_argument("--allow-duplicates", dest="allow_duplicates",
                        action="store_true",
                        help="allow duplicate identifiers. Further occurances "
                        "of an identifier are suffixed by an '_\\%i' ")

    parser.add_argument("--regex-identifier", dest="regex_identifier",
                        type=str,
                        help="regular expression for extracting the "
                        "identifier from fasta description line.")

    parser.add_argument("--force-output", dest="force", action="store_true",
                        help="force overwriting of existing files ")

    translator_choices = ("solexa", "phred", "bytes", "range200")
    parser.add_argument("-t", "--translator", dest="translator", type=str,
                        choices=translator_choices,
                        help="translate numerical quality scores. "
                        "Valid choices are \\%s." %
                        ", ".join(translator_choices))

    group = parser.add_argument_group("Compression options")

    compression_choices = ("lzo", "zlib", "gzip", "dictzip", "bzip2", "debug")
    group.add_argument("-c", "--compression", dest="compression", type=str,
                       choices=compression_choices,
                       help="compress database, using specified compression "
                       "method. "
                       "Valid choices are %s, but depend on availability on the "
                       "system " % ", ".join(compression_choices))

    group.add_argument("--random-access-points", dest="random_access_points",
                       type=int,
                       help="set random access points every # number "
                       "of nucleotides for block compression schemes ")

    group.add_argument(
        "--compress-index", dest="compress_index",
        action="store_true",
        help="compress index. The default is to use a plain-text, "
        "human-readable index.")

    parser.add_argument_group(group)

    parser.set_defaults(
        extract=None,
        input_format="zero-both-open",
        benchmark_fragment_size=1000,
        benchmark_num_iterations=1000000,
        benchmark=False,
        compression=None,
        random_access_points=0,
        synonyms=None,
        verify=None,
        verify_num_iterations=100000,
        verify_fragment_size=100,
        clean_sequence=False,
        allow_duplicates=False,
        regex_identifier=None,
        compress_index=False,
        file_format="auto",
        force=False,
        translator=None)

    (args, unknown) = E.start(parser,
                              unknowns=True)

    if args.synonyms:
        synonyms = {}
        for x in args.synonyms.split(","):
            a, b = x.split("=")
            a = a.strip()
            b = b.strip()
            if a not in synonyms:
                synonyms[a] = []
            synonyms[a].append(b)
    else:
        synonyms = None

    if args.translator:
        if args.translator == "phred":
            args.translator = IndexedFasta.TranslatorPhred()
        elif args.translator == "solexa":
            args.translator = IndexedFasta.TranslatorSolexa()
        elif args.translator == "bytes":
            args.translator = IndexedFasta.TranslatorBytes()
        elif args.translator == "range200":
            args.translator = IndexedFasta.TranslatorRange200()
        else:
            raise ValueError("unknown translator %s" % args.translator)

    if args.extract:
        fasta = IndexedFasta.IndexedFasta(unknown[0])
        fasta.setTranslator(args.translator)
        converter = IndexedFasta.getConverter(args.input_format)

        contig, strand, start, end = IndexedFasta.parseCoordinates(
            args.extract)
        sequence = fasta.getSequence(contig, strand,
                                     start, end,
                                     converter=converter)
        args.stdout.write(">%s\n%s\n" %
                          (args.extract, sequence))

    elif args.benchmark:
        import timeit
        timer = timeit.Timer(
            stmt="IndexedFasta.benchmarkRandomFragment(fasta=fasta, size=%i)" %
            (args.benchmark_fragment_size),
            setup="from cgat import IndexedFasta\n"
            "fasta=IndexedFasta.IndexedFasta('%s')" % (unknown[0]))

        t = timer.timeit(number=args.benchmark_num_iterations)
        args.stdout.write("iter\tsize\ttime\n")
        args.stdout.write("%i\t%i\t%i\n" % (
            args.benchmark_num_iterations,
            args.benchmark_fragment_size, t))

    elif args.verify:
        fasta1 = IndexedFasta.IndexedFasta(unknown[0])
        fasta2 = IndexedFasta.IndexedFasta(args.verify)
        nerrors1 = IndexedFasta.verify(fasta1, fasta2,
                                       args.verify_num_iterations,
                                       args.verify_fragment_size,
                                       stdout=args.stdout)
        args.stdout.write("errors=%i\n" % (nerrors1))
        nerrors2 = IndexedFasta.verify(fasta2, fasta1,
                                       args.verify_num_iterations,
                                       args.verify_fragment_size,
                                       stdout=args.stdout)
        args.stdout.write("errors=%i\n" % (nerrors2))
    elif args.compress_index:
        fasta = IndexedFasta.IndexedFasta(unknown[0])
        fasta.compressIndex()
    else:
        if args.loglevel >= 1:
            args.stdlog.write("# creating database %s\n" % unknown[0])
            args.stdlog.write("# indexing the following files: \n# %s\n" %
                              (" \n# ".join(unknown[1:])))
            args.stdlog.flush()

            if synonyms:
                args.stdlog.write("# Applying the following synonyms:\n")
                for k, v in list(synonyms.items()):
                    args.stdlog.write("# %s=%s\n" % (k, ",".join(v)))
                args.stdlog.flush()
        if len(unknown) < 2:
            print(globals()["__doc__"])
            sys.exit(1)

        iterator = IndexedFasta.MultipleFastaIterator(
            unknown[1:],
            regex_identifier=args.regex_identifier,
            format=args.file_format)

        IndexedFasta.createDatabase(
            unknown[0],
            iterator,
            synonyms=synonyms,
            random_access_points=args.random_access_points,
            compression=args.compression,
            clean_sequence=args.clean_sequence,
            allow_duplicates=args.allow_duplicates,
            translator=args.translator,
            force=args.force)

    E.stop()


if __name__ == "__main__":
    sys.exit(main())



'''
fastq2fastq.py - manipulate fastq files
=============================================

:Tags: Genomics NGS Sequences FASTQ Manipulation

Purpose
-------

This script performs manipulations on :term:`fastq` formatted
files. For example it can be used to change the quality score format
or sample a subset of reads.

The script predominantly is used for manipulation of single fastq
files. However, for some of its functionality it will take paired data
using the ``--pair-fastq-file`` and ``--output-filename-pattern`` options.
This applies to the ``sample`` and ``sort`` methods.

Usage
-----

Example::
  In this example we randomly sample 50% of reads from paired data provided in
  two :term:`fastq` files.

   head in.fastq.1

   @SRR111956.1 HWUSI-EAS618:7:1:27:1582 length=36
   CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
   +SRR111956.1 HWUSI-EAS618:7:1:27:1582 length=36
   =@A@9@BAB@;@BABA?=;@@BB<A@9@;@2>@;??
   @SRR111956.2 HWUSI-EAS618:7:1:29:1664 length=36
   CCCCCCCCCCCCCCCCCCCCCCCCCCCACCCCCCCC
   +SRR111956.2 HWUSI-EAS618:7:1:29:1664 length=36
   =B@9@0>A<B=B=AAA?;*(@A>(@<=*9=9@BA>7
   @SRR111956.3 HWUSI-EAS618:7:1:38:878 length=36
   AGTGAGCAGGGAAACAATGTCTGTCTAAGAATTTGA

   head in.fastq.2

   +SRR111956.3 HWUSI-EAS618:7:1:38:878 length=36
   <?@BA?;A=@BA>;@@7###################
   @SRR111956.4 HWUSI-EAS618:7:1:38:1783 length=36
   ATTAGTATTATCCATTTATATAATCAATAAAAATGT
   +SRR111956.4 HWUSI-EAS618:7:1:38:1783 length=36
   ?ABBA2CCBBB2?=BB@C>=AAC@A=CBB#######
   @SRR111956.5 HWUSI-EAS618:7:1:39:1305 length=36
   CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
   +SRR111956.5 HWUSI-EAS618:7:1:39:1305 length=36
   AA>5;A>*91?=AAA@@BBA<B=?ABA>2>?A<BB@

   command-line::
     cat in.fastq.1 | python fastq2fastq.py
                      --method=sample --sample-size 0.5
                      --pair-fastq-file in.fastq.2
                      --output-filename-pattern out.fastq.2
                      > out.fastq.1

   head out.fastq.1
   @SRR111956.1 HWUSI-EAS618:7:1:27:1582 length=36
   CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
   +
   =@A@9@BAB@;@BABA?=;@@BB<A@9@;@2>@;??
   @SRR111956.2 HWUSI-EAS618:7:1:29:1664 length=36
   CCCCCCCCCCCCCCCCCCCCCCCCCCCACCCCCCCC
   +
   =B@9@0>A<B=B=AAA?;*(@A>(@<=*9=9@BA>7
   @SRR111956.3 HWUSI-EAS618:7:1:38:878 length=36
   AGTGAGCAGGGAAACAATGTCTGTCTAAGAATTTGA
   +
   <?@BA?;A=@BA>;@@7###################
   @SRR111956.4 HWUSI-EAS618:7:1:38:1783 length=36
   ATTAGTATTATCCATTTATATAATCAATAAAAATGT
   +
   ?ABBA2CCBBB2?=BB@C>=AAC@A=CBB#######

Options
-------

The following methods are implemented (``--method``).

``change-format``

    change the quality format to new format given as
    target-format. Options are ``sanger``,
  ``solexa``, ``phred64``, ``integer`` and ``illumina-1.8``

``sample``

    Sub-sample a fastq file. The size of the sample is set by
    --sample-size

``unique``

    Remove duplicate reads based on read name

``trim3``

    Trim a fixed number of nucleotides from the 3' end of reads.
    (see ``--num-bases``). Note that there are better tools for
   trimming.

``trim5``

    Trim a fixed number of nucleotides from the 5' end of reads.
    (see ``--num-bases``). Note that there are better tools for
   trimming.

``sort``

    Sort the fastq file by read name.

``renumber-reads``

    Rename the reads based on pattern given in ``--pattern-identifier``
    e.g. ``--pattern-identifier="read_%010i"``

Type::

   python fastq2fastq.py --help

for command line help.


Command line options
--------------------

'''
import collections
import sys
import os
import re
import random
import pysam
import numpy
import cgatcore.experiment as E
import cgatcore.iotools as iotools
import cgat.Fastq as Fastq
import cgat.Genomics as Genomics


def process_cgat(options):

    c = E.Counter()

    assert options.input_fastq_file == "-"

    if options.method == "change-format":
        for record in Fastq.iterate_convert(options.stdin,
                                            format=options.target_format,
                                            guess=options.guess_format):
            c.input += 1
            options.stdout.write("%s\n" % record)
            c.output += 1

    elif options.method == "grep":
        for record in Fastq.iterate(options.stdin):
            if re.match(options.grep_pattern, record.seq):
                options.stdout.write("%s\n" % record)

    elif options.method == "reverse-complement":
        for record in Fastq.iterate(options.stdin):
            record.seq = Genomics.complement(record.seq)
            record.quals = record.quals[::-1]
            options.stdout.write("%s\n" % record)

    elif options.method == "sample":
        sample_threshold = min(1.0, options.sample_size)

        random.seed(options.seed)

        if options.pair:
            if not options.output_filename_pattern:
                raise ValueError(
                    "please specify output filename pattern for "
                    "second pair (--output-filename-pattern)")

            outfile1 = options.stdout
            outfile2 = iotools.open_file(options.output_filename_pattern, "w")

            for record1, record2 in zip(
                    Fastq.iterate(options.stdin),
                    Fastq.iterate(iotools.open_file(options.pair))):
                c.input += 1
                if random.random() <= sample_threshold:
                    c.output += 1
                    outfile1.write("%s\n" % record1)
                    outfile2.write("%s\n" % record2)
        else:
            for record in Fastq.iterate(options.stdin):
                c.input += 1
                if random.random() <= sample_threshold:
                    c.output += 1
                    options.stdout.write("%s\n" % record)

    elif options.method == "apply":
        ids = set(iotools.read_list(iotools.open_file(options.apply)))

        for record in Fastq.iterate(options.stdin):
            c.input += 1
            if re.sub(" .*", "", record.identifier).strip() in ids:
                c.output += 1
                options.stdout.write("%s\n" % record)

    elif options.method == "trim3":
        trim3 = options.nbases
        for record in Fastq.iterate(options.stdin):
            c.input += 1
            record.trim(trim3)
            options.stdout.write("%s\n" % record)
            c.output += 1

    elif options.method == "trim5":
        trim5 = options.nbases
        for record in Fastq.iterate(options.stdin):
            c.input += 1
            record.trim5(trim5)
            options.stdout.write("%s\n" % record)
            c.output += 1

    elif options.method == "unique":
        keys = set()
        for record in Fastq.iterate(options.stdin):
            c.input += 1
            if record.identifier in keys:
                continue
            else:
                keys.add(record.identifier)
            options.stdout.write("%s\n" % record)
            c.output += 1

    # Need to change this to incorporate both pairs
    elif options.method == "sort":
        if not options.pair:
            # This is quicker for a single fastq file
            statement = "paste - - - - | sort -k1,1 -t ' ' | tr '\t' '\n'"
            os.system(statement)
        else:
            if not options.output_filename_pattern:
                raise ValueError(
                    "please specify output filename for second pair "
                    "(--output-filename-pattern)")
            E.warn(
                "consider sorting individual fastq files - "
                "this is memory intensive")
            entries1 = {}
            entries2 = {}

            for record1, record2 in zip(
                    Fastq.iterate(options.stdin),
                    Fastq.iterate(iotools.open_file(options.pair))):
                entries1[
                    record1.identifier[:-2]] = (record1.seq, record1.quals)
                entries2[
                    record2.identifier[:-2]] = (record2.seq, record2.quals)

            outfile1 = options.stdout
            outfile2 = iotools.open_file(options.output_filename_pattern, "w")
            assert len(set(entries1.keys()).intersection(
                set(entries2.keys()))) == len(entries1),\
                "paired files do not contain the same reads "\
                "need to reconcile files"

            for entry in sorted(entries1):
                outfile1.write("@%s/1\n%s\n+\n%s\n" %
                               (entry, entries1[entry][0], entries1[entry][1]))
                outfile2.write("@%s/2\n%s\n+\n%s\n" %
                               (entry, entries2[entry][0], entries2[entry][1]))

    elif options.method == "renumber-reads":
        id_count = 1
        for record in Fastq.iterate(options.stdin):
            record.identifier = options.renumber_pattern % id_count
            id_count += 1
            options.stdout.write("@%s\n%s\n+\n%s\n" %
                                 (record.identifier, record.seq, record.quals))
    return c


def process_daisy(options):

    filter_n = "filter-N" in options.methods

    filter_ont = "filter-ONT" in options.methods

    if "filter-identifier" in options.methods:
        if options.input_filter_tsv is None:
            raise ValueError("please set --input-filter-tsv for method filter-identifier")
        with iotools.open_file(options.input_filter_tsv) as inf:
            filter_identifier = set([x.split()[0].strip() for x in inf.readlines()])
    else:
        filter_identifier = False

    if options.output_removed_tsv:
        outf_removed_tsv = iotools.open_file(options.output_removed_tsv, "w")
    else:
        outf_removed_tsv = None

    if options.output_removed_fastq:
        outf_removed_fastq = iotools.open_file(options.output_removed_fastq, "w")
    else:
        outf_removed_fastq = None

    if options.set_prefix:
        prefix = "{}".format(options.set_prefix)
    else:
        prefix = None

    quality_offset = options.quality_offset
    counter = E.Counter()

    with pysam.FastxFile(options.input_fastq_file) as inf:
        for read in inf:
            counter.input += 1
            remove = False
            if filter_n:
                chars = collections.Counter(read.sequence)
                if "N" in chars and \
                   100.0 * chars["N"] / len(read.sequence) > options.max_percent_N:
                    remove = True
                    counter.filter_n += 1

            if filter_identifier:
                if read.name not in filter_identifier:
                    counter.filter_identifier += 1
                    remove = True

            if filter_ont:
                quals = read.get_quality_array()
                n = len(quals)
                if n < options.min_sequence_length or \
                        float(sum(quals)) / n < options.min_average_quality:
                    counter.remove_ont += 1
                    remove = True

            if remove:
                counter.removed += 1
                if outf_removed_tsv:
                    outf_removed_tsv.write(read.name + "\n")
                if outf_removed_fastq:
                    outf_removed_fastq.write(str(read) + "\n")
                continue

            if prefix:
                read.name = prefix + read.name[2:]

            if quality_offset:
                quals = numpy.array(read.get_quality_array())
                quals += quality_offset
                quals[quals < 0] = 0
                quals += 33
                # pysam fastq is read-only, so fudge it:
                # Note: not outputting description
                read = "@{}\n{}\n+\n{}".format(
                    read.name,
                    read.sequence,
                    "".join([chr(x) for x in quals]))

            counter.output += 1

            options.stdout.write(str(read) + "\n")

    if outf_removed_tsv:
        outf_removed_tsv.close()

    if outf_removed_fastq:
        outf_removed_fastq.close()

    if options.output_stats_tsv:
        with iotools.open_file(options.output_stats_tsv, "w") as outf:
            outf.write(counter.asTable(as_rows=False) + "\n")

    return counter


def main(argv=sys.argv):

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument(
        "-i", "--input-fastq-file", dest="input_fastq_file", type=str,
        help="input fastq file. "
        )

    parser.add_argument(
        "--output-removed-tsv", dest="output_removed_tsv", type=str,
        help="if given, sequence identifiers of removed sequences will "
        "be stored in this file ")

    parser.add_argument(
        "--output-stats-tsv", dest="output_stats_tsv", type=str,
        help="if given, output statistics will be written to this file. "
        )

    parser.add_argument(
        "--output-removed-fastq", dest="output_removed_fastq", type=str,
        help="if given, removed fastq records will "
        "be stored in this file ")

    parser.add_argument(
        "-m", "--method", dest="methods", action="append", type=str,
        choices=("filter-N",
                 "filter-identifier",
                 "filter-ONT",
                 "offset-quality",
                 "apply",
                 "change-format",
                 "renumber-reads",
                 "sample",
                 "sort",
                 "trim3",
                 "trim5",
                 "unique",
                 "reverse-complement",
                 "grep"),
        help="methods to apply ")

    parser.add_argument(
        "--set-prefix", dest="set_prefix", type=str,
        help="set sequence prefix ")

    parser.add_argument(
        "--input-filter-tsv", dest="input_filter_tsv", type=str,
        help="list of sequence ides to filter ")

    parser.add_argument(
        "--min-average-quality", dest="min_average_quality", type=float,
        help="minimum average quality ")

    parser.add_argument(
        "--min-sequence-length", dest="min_sequence_length", type=int,
        help="minimum sequence length ")

    parser.add_argument(
        "--quality-offset", dest="quality_offset", type=int,
        help="offset to modify quality values with ")

    parser.add_argument(
        "--target-format", dest="target_format", type=str,
        choices=('sanger', 'solexa', 'phred64', 'integer', 'illumina-1.8'),
        help="guess quality score format and set quality scores "
        "to format.")

    parser.add_argument(
        "--guess-format", dest="guess_format", type=str,
        choices=('sanger', 'solexa', 'phred64', 'integer', 'illumina-1.8'),
        help="quality score format to assume if ambiguous.")

    parser.add_argument(
        "--sample-size", dest="sample_size", type=float,
        help="proportion of reads to sample. "
        "Provide a proportion of reads to sample, e.g. 0.1 for 10%, "
        "0.5 for 50%, etc.")

    parser.add_argument(
        "--pair-fastq-file", dest="pair", type=str,
        help="if data is paired, filename with second pair. "
        "Implemented for sampling.")

    parser.add_argument(
        "--map-tsv-file", dest="map_tsv_file", type=str,
        help="filename with tab-separated identifiers mapping for "
        "method apply.")

    parser.add_argument(
        "--num-bases", dest="nbases", type=int,
        help="number of bases to trim.")

    parser.add_argument(
        "--seed", dest="seed", type=int,
        help="seed for random number generator.")

    parser.add_argument(
        "--pattern-identifier", dest="renumber_pattern", type=str,
        help="rename reads in file by pattern")

    parser.add_argument(
        "--grep-pattern", dest="grep_pattern", type=str,
        help="subset to reads matching pattern")

    parser.set_defaults(
        input_fastq_file="-",
        methods=[],
        change_format=None,
        guess_format=None,
        sample_size=0.1,
        nbases=0,
        pair=None,
        apply=None,
        seed=None,
        renumber_pattern="read_%010i",
        grep_pattern=".*",
        max_percent_N=10.0,
        set_prefix=None,
        output_removed_tsv=None,
        output_removed_fastq=None,
        output_stats_tsv=None,
        input_filter_tsv=None,
        min_average_quality=0,
        min_sequence_length=0,
        quality_offset=0,
    )

    (args, unknown) = E.start(parser,
                              argv,
                              add_output_options=True,
                              unknowns=True)

    if len(unknown) == 1:
        args.input_fastq_file = unknown[0]

    if len(args.methods) == 0:
        raise ValueError("no method specified, please use --method")

    # this script combines two scripts with different functionalities
    # TODO: to be sanitized
    if args.methods[0] in ["apply",
                           "change-format",
                           "renumber-reads",
                           "sample",
                           "sort",
                           "trim3",
                           "trim5",
                           "unique",
                           "reverse-complement",
                           "grep"]:
        args.method = args.methods[0]
        counter = process_cgat(args)
    else:
        counter = process_daisy(args)

    E.info(counter)
    E.stop()



'''bam2bam.py - modify bam files
=============================

Purpose
-------

This script reads a :term:`bam` formatted file from stdin, performs an
action (see methods below) then outputs a modified :term:`bam`
formatted file on stdout.

.. note::
   You need to redirect logging information to a file (via -L) or turn it off
   via -v 0 in order to get a valid sam/bam file.

Documentation
-------------

The script implements the following methods:

``set-nh``

   set the NH flag. Some tools (bowtie, bwa) do not set the NH flag.
   If set, this option will set the NH flag (for mapped reads).
   This option requires the bam/sam file to be sorted by read name.

``unset-unmapped_mapq``

   some tools set the mapping quality of unmapped reads. This
   causes a violation in the Picard tools.

``filter``

   remove alignments based on a variety of flags. The filtering method
   is determined by the ``--filter-method`` option. These may be
   ``unique``, ``non-unique``, ``mapped``, ``NM`` or ``CM``.  If
   ``unique`` is set, only uniquely mapping reads will be output. If
   ``non-unique`` is set then only multi-mapping reads will be
   output. This method first checks for the NH flag - if set, a unique
   match should have at most NH=1 hits.  If not set, the method checks
   for BWA flags. Currently it checks if X0 is set (X0=Number of best
   hits found by BWA).  If ``mapped`` is given, unmapped reads will be
   removed. If ``NM`` or ``CM`` is set, the alignment of reads in two
   sam files (input and reference) is compared and only reads with a
   lower number of mismatches in the input compared to the reference
   sam file will be kept. If ``CM`` is set, the colourspace mismatch
   tag (for ABI Solid reads) will be used to count differences to the
   reference sam file. By default, the ``NM`` (number of mismatches)
   tag is used. The tag that is used needs to present in both input
   sam file and the reference sam file. If ``unique`` is given this
   wil NOT remove any unmapped reads.  This can be achieved by
   providing the ``filter`` option twice, once each with ``mapped``
   and ``unique``.

   .. note::

      The filter methods can't currently combined with any of
      the other methods - this is work in progress.

``strip-sequence``

   remove the sequence from all reads in a bam-file. Note that
   stripping the sequence will also remove the quality scores.
   Stripping is not reversible if the read names are not unique.

``strip-quality``

   remove the quality scores from all reads in a bam-file.
   Stripping is not reversible if the read names are not unique.

``set-sequence``

   set the sequence and quality scores in the bam file to some dummy
   values ('A' for sequence, 'F' for quality which is a valid score in
   most fastq encodings. Necessary for some tools that can not work
   with bam-files without sequence.

``unstrip``

   add sequence and quality scores back to a bam file. Requires a
   :term:`fastq` formatted file with the sequences and quality scores
   to insert.

``unset-unmapped-mapq``

   sets the mapping quality of unmapped reads to 0.

``keep-first-base``

   keep only the first base of reads so that read counting tools will
   only consider the first base in the counts

``downsample-single``

   generates a downsampled :term:`bam` file by randomly subsampling
   reads from a single ended :term:`bam` file. The downsmpling
   retains multimapping reads. The use of this requires downsampling
   parameter to be set and optionally randomseed.

``downsample-paired``

   generates a downsampled :term:`bam` file by randomly subsampling
   reads from a paired ended :term:`bam` file. The downsampling
   retains multimapping reads. The use of this requires downsampling
   parameter to be set and optionally randomseed.

``add-sequence-error``

   add a certain amount of random error to read sequences. This method
   picks a certain proportion of positions within a read's sequence
   and alters the nucleotide to a randomly chosen alternative. The
   model is naive and applies uniform probabilities for positions and
   nucleotides. The method does not update base qualities, the
   alignment and the NM flag. As a result, error rates that are
   computed via the NM flag will be unaffected. The error rate is set
   by --error-rate.

By default, the script works from stdin and outputs to stdout.

Usage
-----

For example::

   cgat bam2bam --method=filter --filter-method=mapped < in.bam > out.bam

will remove all unmapped reads from the bam-file.

Example for running downsample::

cgat bam2bam --method=downsample-paired --downsample=30000
--randomseed=1 -L out.log < Paired.bam > out.bam

Type::

   cgat bam2bam --help

for command line help.

Command line options
--------------------

'''

import os
import sys
import tempfile
import shutil
import random
import pysam
import cgatcore.experiment as E
import cgatcore.iotools as iotools
import itertools
import math

from cgat.BamTools.bamtools import bam2bam_filter_bam, SetNH


class SubsetBam(object):

    ''' base class for performing downsampling on single and
    paired bam file

    A dictionary of the read names is made and then an
    array matching the numbers of reads in the dictionary will be
    created. This array contains 1s that match the number of reads
    to be downsampled to and 0s to fill out the rest of the array.

    The read is kept if the value in the dictionary matches 1. The
    read is yielded and iterated over to produce the output bam.

    The script will handle multimapping reads.
    '''

    def __init__(self, infile, downsample, paired_end=None,
                 single_end=None, random_seed=None):

        self.pysam_in1, self.pysam_in2 = itertools.tee(infile)
        self.downsample = downsample
        self.paired_end = paired_end
        self.single_end = single_end
        self.random_seed = random_seed

    def list_of_reads(self, paired=None):

        '''
        This will create a dictionary of uniqe reads in the bam
        '''
        read_list = []

        for read in self.pysam_in1:
            if paired is True:
                if read.is_proper_pair:
                    read_list.append(read.qname)
            else:
                read_list.append(read.qname)

        return sorted(set(read_list))

    def downsample_paired(self):

        '''
        This function will downsample a paired bam file.
        It will retain multimapping reads if they have not been
        pre-filtered
        '''
        if self.random_seed is not None:
            random.seed(self.random_seed)

        collect_list = self.list_of_reads(paired=True)
        read_list = random.sample(collect_list, self.downsample)

        if self.downsample == len(collect_list):
            E.warn('''The downsample reads is equal to the
            number of unique reads''')
            for read in self.pysam_in2:
                yield read

        # yield read if it is in read_list, if the read is multimapped
        # then all multimapping reads will be yielded
        else:
            E.warn('''Multimaping reads have been detected and these will
            be output to the final bam file''')
            for read in self.pysam_in2:
                if read.qname in read_list:
                    yield read

    def downsample_single(self):

        '''
        This function will downsample a single bam file.
        It will retain multimapping reads if not pre-filtered
        '''
        if self.random_seed is not None:
            random.seed(self.random_seed)

        collect_list = self.list_of_reads(paired=False)
        read_list = random.sample(collect_list, self.downsample)

        if self.downsample == len(collect_list):
            E.warn('''The downsample reads is equal to the
            number of unique reads''')
            for read in self.pysam_in2:
                yield read

        # yield read if it is in read_list, if the reads is multimapped
        # then all multimapping reads will be yielded
        else:
            E.warn('''Multimaping reads have been detected and these will
            be output to the final bam file''')
            for read in self.pysam_in2:
                if read.qname in read_list:
                    yield read


def process_bam(infile, outfile, options):

    if "filter" in options.methods:
        if "remove-list" in options.filter_methods or "keep-list" in options.filter_methods:

            it = infile.fetch(until_eof=True)
            c = E.Counter()
            if "remove-list" in options.filter_methods:
                for read in it:
                    c.input += 1
                    if read.query_name in filter_query_names:
                        c.skipped += 1
                        continue
                    outfile.write(read)
                    c.output += 1
            elif "keep-list" in options.filter_methods:
                for read in it:
                    c.input += 1
                    if read.query_name not in filter_query_names:
                        c.skipped += 1
                        continue
                    outfile.write(read)
                    c.output += 1

            E.info("category\tcounts\n%s\n" % c.asTable())
        else:
            remove_mismatches, colour_mismatches = False, False

            if "NM" in options.filter_methods:
                remove_mismatches = True

            elif "CM" in options.filter_methods:
                remove_mismatches = True
                colour_mismatches = True

            if "min-length" in options.filter_methods and options.minimum_read_length == 0:
                raise ValueError("please specify --minimum-read-length when using "
                                 "--filter-method=min-read-length")

            if "min-average-base-quality" in options.filter_methods and options.minimum_average_base_quality == 0:
                raise ValueError("please specify --min-average-base-quality when "
                                 "using --filter-method=min-average-base-quality")

            if remove_mismatches:
                if not options.reference_bam:
                    raise ValueError(
                        "requiring reference bam file for removing by "
                        "mismatches")

                pysam_ref = pysam.AlignmentFile(options.reference_bam, "rb")
            else:
                pysam_ref = None

            # filter and flags are the opposite way around
            c = bam2bam_filter_bam(
                infile, outfile, pysam_ref,
                remove_nonunique="unique" in options.filter_methods,
                remove_unique="non-unique" in options.filter_methods,
                remove_contigs=None,
                remove_unmapped="mapped" in options.filter_methods,
                remove_mismatches=remove_mismatches,
                filter_error_rate=options.error_rate,
                colour_mismatches=colour_mismatches,
                minimum_read_length=options.minimum_read_length,
                minimum_average_base_quality=options.minimum_average_base_quality)

            options.stdlog.write("category\tcounts\n%s\n" % c.asTable())
    else:

        # set up the modifying iterators
        it = infile.fetch(until_eof=True)

        def nop(x):
            return None
        # function to check if processing should start
        pre_check_f = nop

        if "unset-unmapped-mapq" in options.methods:
            def unset_unmapped_mapq(i):
                for read in i:
                    if read.is_unmapped:
                        read.mapq = 0
                    yield read
            it = unset_unmapped_mapq(it)

        if "set-sequence" in options.methods:
            def set_sequence(i):
                for read in i:
                    # can't get at length of unmapped reads
                    if read.is_unmapped:
                        read.seq = "A"
                        read.qual = "F"
                    else:
                        read.seq = "A" * read.inferred_length
                        read.qual = "F" * read.inferred_length

                    yield read
            it = set_sequence(it)

        if "strip-sequence" in options.methods or "strip-quality" in \
           options.methods:
            def strip_sequence(i):
                for read in i:
                    read.seq = None
                    yield read

            def check_sequence(reads):
                if reads[0].seq is None:
                    return 'no sequence present'
                return None

            def strip_quality(i):
                for read in i:
                    read.qual = None
                    yield read

            def check_quality(reads):
                if reads[0].qual is None:
                    return 'no quality information present'
                return None

            def strip_match(i):
                for read in i:
                    try:
                        nm = read.opt('NM')
                    except KeyError:
                        nm = 1
                    if nm == 0:
                        read.seq = None
                    yield read

            if options.strip_method == "all":
                if "strip-sequence" in options.methods:
                    it = strip_sequence(it)
                    pre_check_f = check_sequence
                elif "strip-quality" in options.methods:
                    it = strip_quality(it)
                    pre_check_f = check_quality
            elif options.strip_method == "match":
                it = strip_match(it)

        if "unstrip" in options.methods:
            def buildReadDictionary(filename):
                if not os.path.exists(filename):
                    raise OSError("file not found: %s" % filename)
                fastqfile = pysam.FastxFile(filename)
                fastq2sequence = {}
                for x in fastqfile:
                    if x.name in fastq2sequence:
                        raise ValueError(
                            "read %s duplicate - can not unstrip" % x.name)

                    fastq2sequence[x.name] = (x.sequence, x.quality)
                return fastq2sequence

            if not options.fastq_pair1:
                raise ValueError(
                    "please supply fastq file(s) for unstripping")
            fastq2sequence1 = buildReadDictionary(options.fastq_pair1)
            if options.fastq_pair2:
                fastq2sequence2 = buildReadDictionary(options.fastq_pair2)

            def unstrip_unpaired(i):
                for read in i:
                    read.seq, read.qual = fastq2sequence1[read.qname]
                    yield read

            def unstrip_pair(i):
                for read in i:
                    if read.is_read1:
                        read.seq, read.qual = fastq2sequence1[read.qname]
                    else:
                        read.seq, read.qual = fastq2sequence2[read.qname]
                    yield read

            if options.fastq_pair2:
                it = unstrip_pair(it)
            else:
                it = unstrip_unpaired(it)

        if "set-nh" in options.methods:
            it = SetNH(it)

        # keep first base of reads by changing the cigarstring to
        # '1M' and, in reads mapping to the reverse strand,
        # changes the pos to aend - 1
        # Needs to be refactored to make it more general
        # (last base, midpoint, ..)
        if "keep_first_base" in options.methods:
            def keep_first_base(i):
                for read in i:
                    if read.is_reverse:
                        read.pos = read.aend - 1
                        read.cigarstring = '1M'
                    elif not read.is_unmapped:
                        read.cigarstring = '1M'
                    yield read
            it = keep_first_base(it)

        # read first read and check if processing should continue
        # only possible when not working from stdin
        # Refactoring: use cache to also do a pre-check for
        # stdin input.
        if not infile.is_stream:
            # get first read for checking pre-conditions
            first_reads = list(infile.head(1))

            msg = pre_check_f(first_reads)
            if msg is not None:
                if options.force:
                    E.warn('proccessing continues, though: %s' % msg)
                else:
                    E.warn('processing not started: %s' % msg)
                    return

        if "downsample-single" in options.methods:

            if not options.downsample:
                raise ValueError("Please provide downsample size")

            else:
                down = SubsetBam(infile=it,
                                 downsample=options.downsample,
                                 paired_end=None,
                                 single_end=True,
                                 random_seed=options.random_seed)
                it = down.downsample_single()

        if "downsample-paired" in options.methods:

            if not options.downsample:
                raise ValueError("Please provide downsample size")

            else:
                down = SubsetBam(infile=it,
                                 downsample=options.downsample,
                                 paired_end=True,
                                 single_end=None,
                                 random_seed=options.random_seed)
                it = down.downsample_paired()

        if "add-sequence-error" in options.methods:
            def add_sequence_error(i):
                error_rate = options.error_rate
                map_nuc2var = {"A": "CGT",
                               "C": "AGT",
                               "G": "ACT",
                               "T": "ACG"}
                for read in i:
                    sequence = list(read.query_sequence)
                    quals = read.query_qualities
                    npos = int(math.floor(len(sequence) * error_rate))
                    positions = random.sample(range(len(sequence)), npos)
                    for pos in positions:
                        try:
                            alt = map_nuc2var[sequence[pos]]
                        except KeyError:
                            continue
                        sequence[pos] = alt[random.randint(0, len(alt) - 1)]

                    read.query_sequence = "".join(sequence)
                    read.query_qualities = quals
                    yield read

            it = add_sequence_error(it)

        # continue processing till end
        for read in it:
            outfile.write(read)


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument("-m", "--methods", dest="methods", type=str,
                        action="append",
                        choices=("filter",
                                 "keep-first-base",
                                 "set-nh",
                                 "set-sequence",
                                 "strip-sequence",
                                 "strip-quality",
                                 "unstrip",
                                 "unset-unmapped-mapq",
                                 "downsample-single",
                                 "downsample-paired",
                                 "add-sequence-error"),
                        help="methods to apply ")

    parser.add_argument("--strip-method", dest="strip_method", type=str,
                        choices=("all", "match"),
                        help="define which sequences/qualities to strip. "
                        "match means that stripping only applies to entries "
                        "without mismatches (requires NM tag to be present). "
                        )

    parser.add_argument("--filter-method", dest="filter_methods",
                        action="append", type=str,
                        choices=('NM', 'CM',
                                 "mapped", "unique", "non-unique",
                                 "remove-list",
                                 "keep-list",
                                 "error-rate",
                                 "min-read-length",
                                 "min-average-base-quality"),
                        help="filter method to apply to remove alignments "
                        "from a bam file. Multiple methods can be supplied "
                        )

    parser.add_argument("--reference-bam-file", dest="reference_bam",
                        type=str,
                        help="bam-file to filter with ")

    parser.add_argument("--force-output", dest="force", action="store_true",
                        help="force processing. Some methods such "
                        "as strip/unstrip will stop processing if "
                        "they think it not necessary "
                        )

    parser.add_argument("--output-sam", dest="output_sam", action="store_true",
                        help="output in sam format ")

    parser.add_argument(
        "--first-fastq-file", "-1", dest="fastq_pair1", type=str,
        help="fastq file with read information for first "
        "in pair or unpaired. Used for unstripping sequence "
        "and quality scores ")

    parser.add_argument(
        "--second-fastq-file", "-2", dest="fastq_pair2", type=str,
        help="fastq file with read information for second "
        "in pair. Used for unstripping sequence "
        "and quality scores  ")

    parser.add_argument(
        "--downsample", dest="downsample",
        type=int,
        help="Number of reads to downsample to")

    parser.add_argument(
        "--filename-read-list", dest="filename_read_list",
        type=str,
        help="Filename with list of reads to filter if 'keep-list' or 'remove-list' "
        "filter method is chosen ")

    parser.add_argument(
        "--error-rate", dest="error_rate",
        type=float,
        help="error rate to use as filter. Reads with an error rate "
        "higher than the threshold will be removed ")

    parser.add_argument(
        "--minimum-read-length", dest="minimum_read_length",
        type=int,
        help="minimum read length when filtering ")

    parser.add_argument(
        "--minimum-average-base-quality", dest="minimum_average_base_quality",
        type=float,
        help="minimum average base quality when filtering ")

    parser.set_defaults(
        methods=[],
        output_sam=False,
        reference_bam=None,
        filter_methods=[],
        strip_method="all",
        force=False,
        fastq_pair1=None,
        fastq_pair2=None,
        downsample=None,
        random_seed=None,
        filename_read_list=None,
        error_rate=None,
        minimum_read_length=0,
        minimum_average_base_quality=0,
    )

    # add common options (-h/--help, ...) and parse command line
    (args, unknown) = E.start(parser, argv=argv, unknowns=True)

    if args.stdin != sys.stdin:
        bamfile = args.stdin.name
    elif unknown:
        bamfile = unknown[0]
        if len(unknown) > 1:
            raise ValueError("multiple bam files provided in arguments")
    else:
        bamfile = "-"

    if "remove-list" in args.filter_methods or "keep-list" in args.filter_methods:
        if "remove-list" in args.filter_methods and "keep-list" in args.filter_methods:
            raise ValueError("it is not possible to specify remove-list and keep-list")

        with iotools.open_file(args.filename_read_list) as inf:
            filter_query_names = set([x.strip() for x in inf.readlines() if not x.startswith("#")])
        E.info("read query_sequence filter list with {} read names".format(len(filter_query_names)))

    if "error-rate" in args.filter_methods and not args.error_rate:
        raise ValueError("filtering by error-rate requires --error-rate to be set")

    if "add-sequence-error" in args.methods and not args.error_rate:
        raise ValueError("--add-error-rate requires --error-rate to be set")

    E.info('processing %s' % bamfile)
    if bamfile != "-" and iotools.is_empty(bamfile):
        E.warn('ignoring empty file %s' % bamfile)
        E.stop()
        return

    if args.stdout != sys.stdout:
        output_bamfile = args.stdout.name
    else:
        output_bamfile = "-"
        if args.stdlog == sys.stdout:
            raise ValueError("redirect log-stream to file (--log) if outputting to stdout")

    if args.output_sam:
        output_mode = "wh"
    else:
        output_mode = "wb"

    # reading bam from stdin does not work with only the "r" tag
    with pysam.AlignmentFile(bamfile, "rb") as pysam_in:
        with pysam.AlignmentFile(output_bamfile, output_mode,
                                 template=pysam_in) as pysam_out:
            process_bam(pysam_in, pysam_out, args)

    # write footer and output benchmark information.
    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
medip_merge_intervals.py - merge differentially methylated regions
==================================================================

:Tags: Python

Purpose
-------

This script takes the output of DESeq or EdgeR and merges
adjacent intervals that show a similar expression change.

Input is data like this::

   contig start end treatment_name  treatment_mean  treatment_std   control_name    control_mean    control_std     pvalue  qvalue  l2fold  fold    significant     status
   chr1 10000 11000        CD14    32.9785173324   0       CD4     41.7117152603   0       0.199805206526  1.0     0.338926100945  1.26481475319   0       OK
   chr1 14000 15000        CD14    9.32978709019   0       CD4     9.31489982941   0       1.0     1.0     -0.00230390372974       0.998404330063  0       OK
   chr1 15000 16000        CD14    9.04603350905   0       CD4     9.01484414416   0       1.0     1.0     -0.00498279072069       0.996552150193  0       OK
   chr1 16000 17000        CD14    0.457565479197  0       CD4     0.14910378845   0       0.677265200643  1.0     -1.61766129852  0.325863281276  0       OK

The second and third window would be merged, as

1. Their methylation levels are within 10% of each other.
2. They are both not differentially methylated.

It aggregates the following:

* mean values: average
* std values: max
* pvalue: max
* qvalue: max
* fold: min/max (depending on enrichment/depletion)
* l2fold: min/max (depending on enrichment/depletion)

The analysis outputs bed files with intervals that are
potentially activated in one of the conditions. Windows
with a positive fold change are collected in the ``treatment``,
while windows with a negative fold change are collected in the
``control``.

For methylation analysis, it might be more interesting
to report windows that are depleted (instead of enriched)
of signal. Thus, if the option ``--invert`` is given,
windows with a negative l2fold change are labeled ``treatment``.
Less methylation means that this region is "active" in the
``treatment`` condition.

Note that the input is assumed to be sorted by coordinate.

Usage
-----

Example::

   python cgat_script_template.py --help

Type::

   python cgat_script_template.py --help

for command line help.


Command line options
--------------------

'''

import sys
import re
import collections

import cgatcore.experiment as E
import cgatcore.iotools as iotools

DATA = collections.namedtuple(
    "DATA",
    "test_id contig start end treatment_name  treatment_mean  treatment_std "
    "control_name    control_mean    control_std     pvalue  qvalue  "
    "l2fold  fold    significant     status nintervals")


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument("-o", "--min-overlap", dest="min_overlap", type=int,
                        help="minimum overlap")

    parser.add_argument(
        "-w", "--pattern-window",
        dest="pattern_window", type=str,
        help="regular expression to extract window coordinates from "
        "test id ")

    parser.add_argument(
        "-i", "--invert", dest="invert", action="store_true",
        help="invert direction of fold change ")

    parser.set_defaults(min_overlap=10,
                        invert=False,
                        pattern_window="(\S+):(\d+)-(\d+)"),

    # add common options (-h/--help, ...) and parse command line
    (args) = E.start(parser, argv=argv, add_output_options=True)

    outfiles = iotools.FilePool(args.output_filename_pattern)

    if args.invert:
        test_f = lambda l2fold: l2fold < 0
    else:
        test_f = lambda l2fold: l2fold > 0

    def read():

        rx_window = re.compile(args.pattern_window)
        # filter any of the DESeq/EdgeR message that end up at the top of the
        # output file

        for data in iotools.iterate(args.stdin):

            contig, start, end = rx_window.match(data.test_id).groups()
            start, end = list(map(int, (start, end)))

            yield DATA._make((data.test_id,
                              contig, start, end,
                              data.treatment_name,
                              float(data.treatment_mean),
                              float(data.treatment_std),
                              data.control_name,
                              float(data.control_mean),
                              float(data.control_std),
                              float(data.pvalue),
                              float(data.qvalue),
                              float(data.l2fold),
                              float(data.fold),
                              int(data.significant),
                              data.status,
                              0))

    def grouper(data, distance=10):

        last = next(data)
        entries = [last]

        while 1:
            d = next(data)
            if d is None:
                break
            if d.contig == last.contig and d.start < last.start:
                raise ValueError("error not sorted by start")

            if ((d.contig != last.contig) or
                    (d.start - last.end > distance) or
                    (d.status != last.status) or
                    (d.significant != last.significant) or
                    (d.l2fold * last.l2fold < 0)):
                yield entries
                entries = []

            entries.append(d)
            last = d

        yield entries

    counter = E.Counter()

    args.stdout.write("\t".join(DATA._fields) + "\n")

    # set of all sample names - used to create empty files
    samples = set()

    # need to sort by coordinate
    all_data = list(read())
    all_data.sort(key=lambda x: (x.contig, x.start))

    group_id = 0

    for group in grouper(iter(all_data), distance=args.min_overlap):
        group_id += 1

        start, end = group[0].start, group[-1].end
        assert start < end, 'start > end: %s' % str(group)
        n = float(len(group))
        counter.input += n

        g = group[0]

        if g.l2fold < 0:
            l2fold = max([x.l2fold for x in group])
            fold = max([x.fold for x in group])
        else:
            l2fold = min([x.l2fold for x in group])
            fold = min([x.fold for x in group])

        outdata = DATA._make((
            str(group_id),
            g.contig, start, end,
            g.treatment_name,
            sum([x.treatment_mean for x in group]) / n,
            max([x.treatment_std for x in group]),
            g.control_name,
            sum([x.control_mean for x in group]) / n,
            max([x.control_std for x in group]),
            max([x.pvalue for x in group]),
            max([x.qvalue for x in group]),
            l2fold,
            fold,
            g.significant,
            g.status,
            int(n)))

        samples.add(g.treatment_name)
        samples.add(g.control_name)
        if g.significant:
            if test_f(g.l2fold):
                # treatment lower methylation than control
                outfiles.write(
                    g.treatment_name, "%s\t%i\t%i\t%i\t%f\n" % (
                        g.contig, g.start, g.end,
                        group_id,
                        sum([x.treatment_mean for x in group]) / n))

            else:
                outfiles.write(
                    g.control_name, "%s\t%i\t%i\t%i\t%f\n" % (
                        g.contig, g.start, g.end,
                        group_id,
                        sum([x.control_mean for x in group]) / n))

        args.stdout.write("\t".join(map(str, outdata)) + "\n")

        counter.output += 1

    # create empty files
    for sample in samples:
        outfiles.write(sample, "")

    outfiles.close()
    E.info("%s" % counter)

    # write footer and output benchmark information.
    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""Utility functions for the bam2stats utility."""

import collections

import pysam
from pysam.libchtslib cimport *
from pysam.libcbcf cimport VariantFile, VariantRecord, VariantRecordSample
from pysam.libcfaidx cimport FastxFile, FastqProxy
from pysam.libctabix cimport TabixFile
from libc.string cimport strchr
from libc.stdint cimport int8_t
from libc.stdio cimport puts, printf
from cpython cimport array as c_array

import cgatcore.experiment as E
from cgat.Genomics import reverse_complement
import numpy
cimport numpy
DTYPE = numpy.uint32
ctypedef numpy.uint32_t DTYPE_t

cdef count_diagonals(sequence,
                     forward_hash,
                     reverse_hash,
                     uint32_t query_sequence_length,
                     uint32_t kmer_size):
    cdef uint32_t x, count
    cdef uint32_t offset = len(sequence)
    cdef uint32_t l = len(sequence) + query_sequence_length
    cdef numpy.ndarray[DTYPE_t, ndim=1] forward_counts = numpy.zeros(l, dtype=DTYPE)
    cdef numpy.ndarray[DTYPE_t, ndim=1] reverse_counts = numpy.zeros(l, dtype=DTYPE)
    
    for x from 0 <= x < len(sequence) - kmer_size:
        kmer = sequence[x: x + kmer_size]
        if kmer in forward_hash:
            for count in forward_hash[kmer]:
                forward_counts[count - x + offset] += 1

        if kmer in reverse_hash:
            for count in reverse_hash[kmer]:
                reverse_counts[count - x + offset] += 1

    return forward_counts.max(), reverse_counts.max()


def filter_by_sequence(
        query_sequence,
        FastxFile in_stream1,
        FastxFile in_stream2,
        outf_matched1,
        outf_matched2,
        outf_unmatched1,
        outf_unmatched2,
        uint32_t kmer_size=10,
        uint32_t min_kmer_matches=20):

    _query_sequence = query_sequence.encode("ascii")
    reverse_sequence = reverse_complement(query_sequence)
    cdef uint32_t query_sequence_length = len(query_sequence)
    cdef uint32_t x
    cdef char * s = _query_sequence

    forward_hash = collections.defaultdict(list)
    reverse_hash = collections.defaultdict(list)
    for x from 0 <= x < query_sequence_length - kmer_size:
        forward_hash[query_sequence[x:x + kmer_size]].append(x)
        reverse_hash[reverse_sequence[x: x + kmer_size]].append(x)

    cdef FastqProxy read1, read2
    cdef uint32_t ninput = 0
    cdef uint32_t nmatched = 0
    cdef uint32_t nunmatched = 0
    cdef uint32_t fc1, rc1, fc2, rc2

    for read1, read2 in zip(in_stream1, in_stream2):
    
        fc1, rc1 = count_diagonals(
            read1.sequence,
            forward_hash,
            reverse_hash,
            query_sequence_length,
            kmer_size)
        fc2, rc2 = count_diagonals(
            read2.sequence,
            forward_hash,
            reverse_hash,
            query_sequence_length,
            kmer_size)

        if max(fc1, rc1, fc2, rc2) > min_kmer_matches:
            nmatched += 1
            outf_matched1.write(str(read1) + "\n")
            outf_matched2.write(str(read2) + "\n")
        else:
            nunmatched += 1
            outf_unmatched1.write(str(read1) + "\n")
            outf_unmatched2.write(str(read2) + "\n")
        
        ninput += 1
        if ninput % 1000 == 0:
            E.info("iteration: {}, matched={}, unmatched={}, permille_matched={}".format(
                ninput, nmatched, nunmatched, 1000.0 * nmatched / ninput))

    c = E.Counter()
    c.input = ninput
    c.matched = nmatched
    c.unmatched = nunmatched
    return c



'''
cgat_fasta2cDNA.py - converting multi-fasta of exon features into a multi-fasta of spliced cDNAs/RNAs
======================================================================================================

:Tags: Python

Purpose
-------

Usage
-----

.. Example use case

Example::

   python cgat_fasta2cDNA.py

Type::

   python cgat_fasta2cDNA.py --help

for command line help.

Command line options
--------------------

'''

import sys
import cgatcore.experiment as E
import cgatcore.iotools as iotools


def makeSplicedFasta(infile):
    '''
    Merge fasta sequences together into a single
    spliced transcript sequence
    '''

    fasta_dict = {}
    with iotools.open_file(infile) as fafile:
        for line in fafile.readlines():
            if line[0] == '>':
                header = line.rstrip("\n")
                fasta_dict[header] = ''
            else:
                fasta_dict[header] += line.rstrip("\n")

    for key, value in sorted(fasta_dict.items()):
        yield "%s\n%s\n" % (key, value)


def main(argv=None):
    """script main.
    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    # add common options (-h/--help, ...) and parse command line
    (args) = E.start(parser, argv=argv)

    infile = argv[-1]
    for record in makeSplicedFasta(infile):
        options.stdout.write(record)

    # write footer and output benchmark information.
    E.stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""Utility functions for the bam2stats utility."""

import pysam
from pysam.libchtslib cimport *
from pysam.libcbcf cimport VariantFile, VariantRecord, VariantRecordSample
from pysam.libcfaidx cimport FastaFile
from pysam.libctabix cimport TabixFile
from libc.string cimport strchr
from libc.stdint cimport int8_t
from libc.stdio cimport puts, printf
from cpython cimport array as c_array
import array
import base64
import collections
import collections
import hashlib
import itertools
import math
import numpy
import pandas
import re
import string
import sys
import cgatcore.experiment as E


# deprecated: avoid R dependencies in cgat-apps
# import rpy2.robjects
# from rpy2.robjects import r as R
# from rpy2.robjects.packages import importr
# from rpy2.robjects import pandas2ri

cimport numpy


def generate_from_bed(vcf_file, bed_file):
    for bed in bed_file.fetch(parser=pysam.asBed()):
        for v in vcf_file.fetch(bed.contig, bed.start, bed.end):
            yield v

def generate_from_vcf(vcf_file):
    for v in vcf_file:
        yield v

def generate_from_region(vcf_file, region):
    for v in vcf_file.fetch(region=region):
        yield v


ACGT = str.maketrans("ACGT", "TGCA")


cdef class Counter(object):

    cdef int nsamples
    cdef object samples
    cdef bint only_variant_positions

    def __init__(self, samples, only_variant_positions=False):
        self.nsamples = len(samples)
        self.samples = samples
        self.only_variant_positions = only_variant_positions

    cdef process_record(self, VariantRecord record, bint is_snp):
        raise NotImplementedError("base class must implement process_record")

    
cdef class CounterKinship(Counter):

    cdef int32_t * data_genotype_ptr

    cdef numpy.ndarray n_Aa
    cdef numpy.ndarray n_AAaa
    cdef numpy.ndarray n_AaAa
    cdef numpy.ndarray genotype_values
    
    def __init__(self, *args, **kwargs):
        Counter.__init__(self, *args, **kwargs)

        self.data_genotype_ptr = NULL
        self.n_Aa = numpy.zeros(self.nsamples, dtype=numpy.int64)
        # if memory becomes an issue, use triangular matrices
        self.n_AAaa = numpy.zeros((self.nsamples, self.nsamples), dtype=numpy.int64)
        self.n_AaAa = numpy.zeros((self.nsamples, self.nsamples), dtype=numpy.int64)
        self.genotype_values = numpy.zeros(self.nsamples, dtype=numpy.int8)

    cdef process_record(self, VariantRecord record, bint is_snp):

        if not is_snp:
            return

        cdef int mdat = 0
        cdef int32_t * ptr = NULL
        cdef int allele
        cdef bint is_het_i, is_het_j
        cdef int nret
        cdef int _i, _j

        cdef int8_t [:] genotype_values_view = self.genotype_values

        nret = bcf_get_genotypes(
            record.header.ptr,
            record.ptr,
            &self.data_genotype_ptr,
            &mdat)

        if nret == 0:
            raise ValueError("no genotypes")

        nret /= self.nsamples
        ptr = <int32_t*>self.data_genotype_ptr
        for _i from 0 <= _i < self.nsamples:
            # sum up alleles. Missing alleles or more than 2 alleles 
            # will set the value to a large negative value
            genotype_values_view[_i] = 0
            for _j from 0 <= _j < nret:
                if bcf_gt_is_missing(ptr[_j]):
                    genotype_values_view[_i] = -10
                    continue
                if ptr[_j]==bcf_int32_vector_end:
                    break
                allele = bcf_gt_allele(ptr[_j])
                if allele > 1:
                    genotype_values_view[_i] = -10
                genotype_values_view[_i] += allele
            ptr += nret

        for _i from 0 <= _i < self.nsamples:
            if genotype_values_view[_i] < 0:
                continue
            is_het_i = genotype_values_view[_i] == 1
            if is_het_i:
                self.n_Aa[_i] += 1
            for _j from 0 <= _j < _i:
                if genotype_values_view[_j] < 0:
                    continue
                is_het_j = genotype_values_view[_j] == 1
                if is_het_i and is_het_j:
                    self.n_AaAa[_i][_j] += 1
                elif (not is_het_i and
                      not is_het_j and
                      genotype_values_view[_i] != genotype_values_view[_j]):
                    self.n_AAaa[_i][_j] += 1

    def output(self):

        cdef int _i, _j
        with E.open_output_file("kinship") as outf:
            outf.write("sample_i\tsample_j\twithin_kinship\t"
                       "between_kinship\tn_het_i\tn_het_j\tn_homhom\tn_hethet\n")
            for _i from 0 <= _i < self.nsamples:
                for _j from 0 <= _j < _i:
                    try:
                        between_kinship = (self.n_AaAa[_i][_j] - 2.0 * self.n_AAaa[_i][_j]) \
                            / (2.0 * self.n_Aa[_i]) \
                            + 0.5 - 0.25 * (self.n_Aa[_i] + self.n_Aa[_j]) / self.n_Aa[_i]
                        between_kinship_f = "{:6.4f}".format(between_kinship)
                    except ZeroDivisionError:
                        between_kinship_f = ""

                    try:
                        within_kinship = (self.n_AaAa[_i][_j] - 2.0 * self.n_AAaa[_i][_j]) \
                            / (self.n_Aa[_i] + self.n_Aa[_j])
                        within_kinship_f = "{:6.4f}".format(within_kinship)
                    except ZeroDivisionError:
                        within_kinship_f = ""

                    outf.write("\t".join(map(
                        str,
                        (self.samples[_i],
                         self.samples[_j],
                         within_kinship_f,
                         between_kinship_f,
                         self.n_Aa[_i],
                         self.n_Aa[_j],
                         self.n_AAaa[_i][_j],
                         self.n_AaAa[_i][_j]))) + "\n")

    def __dealloc__(self):
        if self.data_genotype_ptr is not NULL:
            free(self.data_genotype_ptr)


cdef class CounterFormatDistributions(Counter):
    
    cdef numpy.ndarray counts
    cdef numpy.ndarray unset_samples
    cdef numpy.ndarray unset_sites
    cdef numpy.ndarray values

    cdef object codes
    cdef int ncodes
    cdef int nbins

    def __init__(self, codes, nbins, *args, **kwargs):
        Counter.__init__(self, *args, **kwargs)

        self.nbins = nbins
        # use byte values to facilitate internal processing
        self.codes = [x.encode("ascii") for x in codes]
        self.ncodes = len(codes)

        self.counts = numpy.zeros(
            (self.ncodes,
             self.nsamples,
             int(nbins) + 1), dtype=numpy.int64)

        self.unset_samples = numpy.zeros(
            (self.ncodes,
             self.nsamples),
            dtype=numpy.int64)

        self.unset_sites = numpy.zeros(
            (self.ncodes,
             self.nsamples + 1),
            dtype=numpy.int64)
        
        self.values = numpy.zeros(
            self.nsamples,
            dtype=numpy.int32)

    cdef process_record(self, VariantRecord record, bint is_snp):

        cdef int64_t [:, :, :] counts_view = self.counts
        cdef int64_t [:, :] unset_samples_view = self.unset_samples
        cdef int64_t [:, :] unset_sites_view = self.unset_sites
        cdef int32_t [:] values_view = self.values
        cdef int32_t * data_ptr = &values_view[0]
        cdef int code_idx, sample_idx, ival
        cdef char * code
        cdef int ndest

        unset = numpy.zeros(self.ncodes, dtype=numpy.int64)
        cdef int64_t [:] unset_view = unset

        # quickly iterate over all values for a particular format
        for code_idx, code in enumerate(self.codes):
            bcf_get_format_int32(
                record.header.ptr,
                record.ptr,
                code,
                &data_ptr,
                &ndest)

            for sample_idx from 0 <= sample_idx < self.nsamples:
                ival = data_ptr[sample_idx]
                if ival == bcf_int32_missing:
                    unset_view[code_idx] += 1
                    unset_samples_view[code_idx][sample_idx] += 1
                elif ival < 0:
                    raise ValueError("received negative value {} for format field {}".format(
                        ival, code))
                else:
                    if ival > self.nbins:
                        ival = self.nbins
                    counts_view[code_idx][sample_idx][ival] += 1

        for code_idx, ival in enumerate(unset_view):
            unset_sites_view[code_idx][ival] += 1

    def output(self):

        str_codes = [x.decode("ascii") for x in self.codes]

        with E.open_output_file("format_per_sample") as outf:
            outf.write("FORMAT\tbin\t{}\n".format(
                "\t".join(self.samples)))

            for code_idx, code in enumerate(str_codes):
                bins = numpy.arange(0, self.nbins + 1, 1)
                bins.shape = (self.nbins + 1, 1)
                mm = self.counts[code_idx].transpose()
                mm = numpy.hstack([bins, mm])
                assert mm.shape == (self.nbins + 1, self.nsamples + 1)

                # some metrics are discretized, so remove any all-0 rows            
                row_sums = mm.sum(axis=1)
                comp = mm[:, 0]
                mm = mm[comp != row_sums]
                for row in mm:
                    outf.write("{}\t{}\n".format(code, "\t".join(map(str, row))))

        with E.open_output_file("format_unset_samples") as outf:
            outf.write("FORMAT\t{}\n".format(
                "\t".join(self.samples)))
            for code_idx, code in enumerate(str_codes):
                outf.write("{}\t{}\n".format(
                    code,
                    "\t".join(map(str, self.unset_samples[code_idx]))))
        bins = numpy.arange(0, self.nsamples + 1)
        bins.shape = (self.nsamples + 1, 1)

        # output as binary for numpy savetxt function
        with E.open_output_file("format_unset_sites", mode="wb") as outf:
            header = "bin\t{}\n".format(
                "\t".join(str_codes))
            outf.write(header.encode("ascii"))
            mm = numpy.hstack([bins, self.unset_sites.transpose()])
            numpy.savetxt(outf,
                          mm,
                          delimiter="\t",
                          fmt="%i")


cdef class CounterMutationalSignature(Counter):

    cdef object signatures
    cdef object profile

    cdef FastaFile fasta_in

    def __init__(self, fasta_in, *args, **kwargs):
        Counter.__init__(self, *args, **kwargs)
        self.fasta_in = fasta_in

        # Mutation profile
        #
        # C -> A   G    T   C>A = C->A, G->T
        # G    T   C -> A
        #
        # C -> G   C    G   C>G = C->G, G->C
        # G    C   G -> C
        #
        # C -> T   G    A   C>T = C->T, G->A
        # G    A   C -> T
        #
        # T -> A   T    A   T>A = T->A, A->T
        # A    T   A -> T
        #
        # T -> C   A    G   T>C = T->C, A->G
        # A    G   T -> C

        # T -> G   A    C   T>G = T->G, A->C
        # A    C   T -> G

        self.profile = {
            "CA": ("C>A", 0),
            "GT": ("C>A", 1),
            "CG": ("C>G", 0),
            "GC": ("C>G", 1),
            "CT": ("C>T", 0),
            "GA": ("C>T", 1),
            "TA": ("T>A", 0),
            "AT": ("T>A", 1),
            "TC": ("T>C", 0),
            "AG": ("T>C", 1),
            "TG": ("T>G", 0),
            "AC": ("T>G", 1)
        }
        assert len(self.profile.keys()) == 12

        self.signatures = collections.defaultdict(
            lambda: collections.defaultdict(
                lambda: collections.defaultdict(int)))

    cdef process_record(self, VariantRecord record, bint is_snp):

        # note: pos in VariantFile is 1-based
        if not is_snp:
            return

        alt = record.alts[0]
        context = self.fasta_in.fetch(record.chrom,
                                      record.pos - 2,
                                      record.pos + 1).upper()

        assert context[1] == record.ref, \
            "reference sequence mismatch? expected {} at {}:{}, got {}".format(
                context[1], record.chrom, record.pos, record.ref)

        # skip any non ACGT bases
        if not(context[0] in "ACGT" and
               context[1] in "ACGT" and
               context[2] in "ACGT"):
            return

        profile_class, reverse = self.profile[record.ref + alt]
        if reverse:
            context = context.translate(ACGT)[::-1]

        c = "{}.{}".format(context[0], context[2])

        all_alleles = collections.defaultdict(int)
        first_variants = None
        is_unique = False
        for s in self.samples:
            ai = list(record.samples[s].allele_indices)
            variants = [x for x in ai if x and x > 0]

            if len(variants) > 0:
                self.signatures[s][profile_class][c] += 1
                for v in variants:
                    all_alleles[v] += 1

            if first_variants is None:
                first_variants = tuple(sorted(variants))
            else:
                if tuple(sorted(variants)) != first_variants:
                    is_unique = True

        if is_unique:
            self.signatures["unique"][profile_class][c] += 1

    def output(self):

        with E.open_output_file("mutation_profile") as outf:
            outf.write("sample\tsignature\tcontext\tcount\t"
                       "percent_sample\tpercent_context\n")
            for s, dd in sorted(self.signatures.items()):
                total_class = collections.defaultdict(int)
                for cls, d in dd.items():
                    t = 0
                    for alt, count in d.items():
                        t += count
                    total_class[cls] = float(t)
                total_sample = sum(total_class.values())

                for cls, d in sorted(dd.items()):
                    for context, count in sorted(d.items()):
                        outf.write("\t".join(map(
                            str,
                            (s, cls, context, count,
                             100.0 * count / total_sample,
                             100.0 * count / total_class[cls])
                        )) + "\n")



cdef class CounterMutationalSignatureProfile(CounterMutationalSignature):
    """compute signutares from mutational profile.

    This module requires the deconstructSigs R library. To install:
    source("https://bioconductor.org/biocLite.R")
    biocLite()
    biocLite(c("BSgenome", "BSgenome.Hsapiens.UCSC.hg19", "GenomeInfoDb"))
    install.packages("deconstructSigs")
    """

    cdef object signatures_database
    cdef object counts_method

    def __init__(self, *args, **kwargs):
        raise NotImplementedError("R dependencies are not resolved")
    #     CounterMutationalSignature.__init__(self, *args, **kwargs)
    #     deconstructSigs = importr('deconstructSigs')
    #     R('''library(deconstructSigs)''')

    #     self.signatures_database = "signatures.cosmic"
    #     self.counts_method = "genome"
        
    # def output(self):

    #     colnames = list(R('''colnames(randomly.generated.tumors)'''))

    #     # colnames are of format:
    #     # [1] "A[C>A]A" "A[C>A]C" "A[C>A]G" "A[C>A]T" "C[C>A]A" "C[C>A]C" "C[C>A]G"
    #     # [8] "C[C>A]T" "G[C>A]A" "G[C>A]C" "G[C>A]G" "G[C>A]T" "T[C>A]A" "T[C>A]C"

    #     rows = []
    #     samples = []
    #     for sample, dd in self.signatures.items():
    #         values = {}
    #         for cls, d in dd.items():
    #             for context, count in d.items():
    #                 # context = "G.G", cls = C>T
    #                 values["{}[{}]{}".format(
    #                     context[0], cls, context[2])] = count

    #         extra = set(values.keys()).difference(colnames)
    #         missing = set(colnames).difference(values.keys())
    #         if len(extra) > 0:
    #             raise AssertionError(
    #                 "unexpected additional columns in sample {}: {}".format(
    #                     sample, sorted(extra)))
    #         if len(missing) > 0:
    #             E.warn("missing counts for sample {}: {} will be set to 0".format(
    #                 sample, sorted(missing)))

    #         samples.append(sample)
    #         rows.append([values.get(x, 0) for x in colnames])

    #     df = pandas.DataFrame.from_records(rows,
    #                                        columns=colnames,
    #                                        index=samples)

    #     # convert to R dataframe. The conversion translates special charactes
    #     # such as []< to ., so set column names explicitely.
    #     rdf = pandas2ri.py2ri(df)
    #     rdf.colnames = rpy2.robjects.StrVector(colnames)
    #     R.assign("rdf", rdf)
    #     results = []
    #     for sample in df.index:
    #         R('''result = whichSignatures(rdf,
    #         sample.id='{sample}',
    #         signatures.ref={signatures},
    #         contexts.needed=TRUE,
    #         signature.cutoff=0,
    #         tri.counts.method="{counts_method}")'''.format(
    #             sample=sample,
    #             signatures=self.signatures_database,
    #             counts_method=self.counts_method))
    #         results.append(R('''result$weights'''))

    #     results = pandas.concat(results, axis=0)
    #     results.columns = [re.sub("[.]", "", x).lower() for x in results.columns]
    #     results.index.name = "sample"
    #     results = results.reset_index()

    #     with E.open_output_file("mutation_profile_signatures") as outf:
    #         results.to_csv(outf, sep="\t", index=False)


cdef class CounterGCContext(Counter):

    cdef FastaFile fasta_in
    cdef int nbins
    cdef int window_size
    cdef numpy.ndarray counts

    def __init__(self, fasta_in, *args, **kwargs):
        Counter.__init__(self, *args, **kwargs)
        self.fasta_in = fasta_in
        self.nbins = 100
        self.counts = numpy.zeros((self.nsamples, self.nbins + 1))
        self.window_size = 50

    cdef process_record(self, VariantRecord record, bint is_snp):

        if not is_snp:
            return
        context = self.fasta_in.fetch(record.chrom,
                                      max(0, record.pos - self.window_size),
                                      record.pos + self.window_size).upper()
        counts = collections.Counter(context)
        gc = counts["G"] + counts["C"]
        at = counts["A"] + counts["T"]
        gc_content = int(math.floor(100.0 * gc / (gc + at)))
        c = "{}.{}".format(context[0], context[2])
        for idx, s in enumerate(self.samples):
            try:
                ai = list(record.samples[s].allele_indices)
            except TypeError:
                # no GT field
                continue

            if self.only_variant_positions:
                if len([x for x in ai if x > 0]):
                    continue

            self.counts[idx][gc_content] += 1

    def output(self):
        with E.open_output_file("gc_context") as outf:
            outf.write("percent_gc\t{}\n".format("\t".join(self.samples) ))
            bins = numpy.arange(0, self.nbins + 1, 1)
            bins.shape = (self.nbins + 1, 1)
            mm = self.counts.transpose()
            mm = numpy.hstack([bins, mm])
            assert mm.shape == (self.nbins + 1, self.nsamples + 1)
            for row in mm:
                outf.write("{}\n".format("\t".join(map(str, row))))


cdef class CounterGCDepthProfile(Counter):

    cdef FastaFile fasta_in
    cdef int nbins_gc
    cdef int nbins_dp
    cdef int window_size
    cdef numpy.ndarray counts

    def __init__(self, fasta_in, gc_window_size=50, *args, **kwargs):
        Counter.__init__(self, *args, **kwargs)
        self.fasta_in = fasta_in
        self.nbins_gc = 100
        self.nbins_dp = 1000
        self.counts = numpy.zeros(
            (self.nsamples, self.nbins_dp + 1, self.nbins_gc + 1))
        self.window_size = gc_window_size

    cdef process_record(self, VariantRecord record, bint is_snp):

        if not is_snp:
            return

        context = self.fasta_in.fetch(record.chrom,
                                      max(0, record.pos - self.window_size),
                                      record.pos + self.window_size).upper()

        counts = collections.Counter(context)
        gc = counts["G"] + counts["C"]
        at = counts["A"] + counts["T"]
        gc_content = int(math.floor(100.0 * gc / (gc + at)))

        c = "{}.{}".format(context[0], context[2])

        for idx, s in enumerate(self.samples):
            try:
                ai = list(record.samples[s].allele_indices)
            except TypeError:
                # no GT field
                continue

            if self.only_variant_positions:
                if len([x for x in ai if x > 0]):
                    continue
                    
            depth = record.samples[s]["DP"]
            if depth != None:
                depth = min(depth, self.nbins_dp)
                self.counts[idx][depth][gc_content] += 1

    def output(self):
        with E.open_output_file("gc_dp_prof") as outf:
            gc_bins = numpy.arange(0, self.nbins_gc + 1, 1)

            dp_bins = numpy.arange(0, self.nbins_dp + 1, 1)
            dp_bins.shape = (self.nbins_dp + 1, 1)

            outf.write("sample\tdp\tbin\t{}\n".format("\t".join(map(str, gc_bins))))

            for sample_idx, sample in enumerate(self.samples):
                mm = self.counts[sample_idx]
                mm = numpy.hstack([dp_bins, mm])
                assert mm.shape == (self.nbins_dp + 1, self.nbins_gc + 2)
                # some metrics are discretized, so remove any all-0 rows
                row_sums = mm.sum(axis=1)
                comp = mm[:, 0]
                mm = mm[comp != row_sums]
                for row in mm:
                    outf.write("{}\t{}\n".format(sample, "\t".join(map(str, row))))

        with E.open_output_file("gc_dp_profile_stats") as outf:
            gc_bins = numpy.arange(0, self.nbins_gc + 1, 1)
            dp_bins = numpy.arange(0, self.nbins_dp + 1, 1)
            dp_bins.shape = (self.nbins_dp + 1, 1)
            outf.write ("sample\tgc_bin\tmean\n")
            for sample_idx, sample in enumerate(self.samples):
                mm = self.counts[sample_idx]
                if mm.sum() == 0:
                    continue

                sums = (dp_bins * mm).sum(axis=0)
                counts = mm.sum(axis=0)
                means = numpy.nan_to_num((sums / counts))

                for gc_bin, mean in zip(gc_bins, means):
                    outf.write("{}\t{}\t{:.2f}\n".format(sample, gc_bin, mean))


def vcf2stats_count(VariantFile vcf_in,
                    FastaFile fasta_in,
                    TabixFile bed_in,
                    options):

    if options.region is not None:
        vcf_records = generate_from_region(vcf_in, options.region)
    elif bed_in is not None:
        vcf_records = generate_from_bed(vcf_in, bed_in)
    else:
        vcf_records = generate_from_vcf(vcf_in)

    samples = list(vcf_in.header.samples)

    counters = []

    for method in options.methods:
        if method == "mutational-signature":
            counters.append(CounterMutationalSignature(
                fasta_in=fasta_in,
                samples=samples))
        elif method == "mutational-signature-profile":
            counters.append(CounterMutationalSignatureProfile(
                fasta_in=fasta_in,
                samples=samples))
        elif method == "kinship":
            counters.append(CounterKinship(
                samples=samples))
        elif method == "format-distribution":
            counters.append(CounterFormatDistributions(
                nbins=options.format_distribution_nbins,
                codes=options.format_distributions,
                samples=samples))
        elif method == "gc-context":
            counters.append(CounterGCContext(
                fasta_in=fasta_in,
                samples=samples))
        elif method == "gc-depth-profile":
            counters.append(CounterGCDepthProfile(
                fasta_in=fasta_in,
                samples=samples,
                gc_window_size=options.gc_window_size,
                only_variant_positions=options.only_variant_positions))

    cdef bint is_snp

    cdef int report_step = options.report_step
    cdef int record_idx
    cdef VariantRecord record

    cdef Counter counter

    for record_idx, record in enumerate(vcf_records):
        if (record_idx % options.report_step) == 0:
            E.debug("iteration {}: {}:{}".format(record_idx, record.chrom, record.pos))

        if record.filter.values()[0].name != "PASS":
            continue

        # skip <NON_REF>
        alts = record.alts
        if alts[0][0] == "<":
            continue

        ref = record.ref
        is_snp = (len(ref) == 1 and
                  len(alts) == 1 and
                  len(alts[0]) == 1)

        for counter in counters:
            counter.process_record(record, is_snp)

    for counter in counters:
        counter.output()



"""modify a fastq file.
"""

import collections
import sys
import pysam
import cgatcore.experiment as E
import cgatcore.iotools as iotools


def main(argv=sys.argv):

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument(
        "-i", "--input-fastq-file", dest="input_fastq_file", type=str,
        help="input fastq file. "
        )

    parser.add_argument(
        "-m", "--method", dest="methods", action="append", type=str,
        choices=("length", ),
        help="methods to apply ")

    parser.set_defaults(
        methods=[],
        input_fastq_file=None,
    )

    (args, unknown) = E.start(parser,
                              argv,
                              unknowns=True)

    if len(unknown) == 1:
        args.input_fastq_file = unknown[0]

    if args.input_fastq_file is None:
        raise ValueError("missing input fastq file")

    counter = E.Counter()

    # note: complete rewrite with Counters, currently only length
    if args.methods != ["length"]:
        raise NotImplementedError()

    with pysam.FastqFile(args.input_fastq_file) as inf:

        for read in inf:
            counter.input += 1
            args.stdout.write("\t".join(
                map(str, (read.name, len(read.sequence)))) + "\n")

            counter.output += 1

    E.info(counter)
    E.stop()



"""Compute statistics from a VCF file.

Implemeted methods
==================

mutation-profile
----------------

Compute a mutation profile for each sample in the VCF file. For each
central base in a tri-nucleotide, the following equivalence classes
are used::

    R    A   R    A   Code     Forward strand substitutions
    ---------------
    C -> A   G    T   C>A    = C->A, G->T
    G    T   C -> A

    C -> G   C    G   C>G    = C->G, G->C
    G    C   G -> C

    C -> T   G    A   C>T    = C->T, G->A
    G    A   C -> T

    T -> A   T    A   T>A    = T->A, A->T
    A    T   A -> T

    T -> C   A    G   T>C    = T->C, A->G
    A    G   T -> C

    T -> G   A    C   T>G    = T->G, A->C
    A    C   T -> G

kinship
-------

Compute kinship coefficient for all pairs of samples in VCF.
The kinship coefficient is estimated using the robust estimator
by Manichaikul et al (2010).

The robust estimator for between family relationship is (eq. 11):

phi_ij = N_{Aa, Aa} - 2 N_{AA,aa} / (2 N_{Aa}(i))
         + 1/2
         - 1/4 * N_{Aa}(i) + N_{Aa}(j) / N_{Aa}(i)

The robust estimator for within family relationship is (eq. 9):

phi_ij = N_{Aa,Aa} - 2 N_{AA,aa} / (N_{Aa}(i) + N_{Aa}(j))

with:

N_{Aa,Aa}: # of variants where both individuals are heterozygous
N_{AA,aa}: # of variants where both individuals are homozygous different
N_{Aa}(i): # of heterozygous variants in individual i

format-distribution
-------------------

Compute distribution of one or more metrics in the format field. This method
outputs several tables:

format_per_sample

   Histogram over the FORMAT field. The first two columns in the table
   are `FORMAT`, the FORMAT field specifier, and `bin`. These columns
   are followed by each sample as a column.

format_unset_samples

   Table showing the number of unset FORMAT fields per sample. The
   first column is FORMAT followed by columns for each sample.

format_unset_sites

   Table showing the distribution of sites for each FORMAT field that
   have no annotation for a particular field. This table has the
   column `bin` followed by one column for each FORMAT field.

gc-depth-profile
----------------

gc-context
----------



"""

import os
import sys
import pysam

import cgatcore.experiment as E

from cgat.VCFTools import vcf2stats_count


def main(argv=sys.argv):

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument(
        "-i", "--input-vcf", dest="input_vcf_file", type=str,
        help="input vcf file")

    parser.add_argument(
        "-f", "--input-fasta", dest="input_fasta_file", type=str,
        help="input fasta file. faidx indexed reference sequence file to "
        "determine INDEL context ")

    parser.add_argument(
        "-e", "--input-bed", dest="input_bed_file", type=str,
        help="input file with intervals. Tab-delimited file of intervals "
        "in bed format to restrict analysis to. ")

    parser.add_argument(
        "-r", "--region", dest="region", type=str,
        help="Region string to restrict analysis to. Takes precedence "
        "over --input-bed. ")

    parser.add_argument(
        "-m", "--method", dest="methods", action="append", type=str,
        choices=("mutational-signature",
                 "mutational-signature-profile",
                 "kinship",
                 "format-distribution",
                 "gc-context",
                 "gc-depth-profile"),
        help="methods to apply ")

    parser.add_argument(
        "--format-distribution", dest="format_distributions", action="append",
        type=str,
        help="format to compute histograms on. Option can specified multiple times. "
        "At the moment, only integer metrics are supported ")

    parser.add_argument(
        "--format-distribution-nbins", dest="format_distributions_nbins", type=int,
        help="number of bins to use for histograms ")

    parser.add_argument(
        "--only-variant-positions", dest="only_variant_positions",
        action="store_true",
        help="only use variant positions ")

    parser.add_argument(
        "--gc-window-size", dest="gc_window_size", type=int,
        help="(half) window size to use for G+C computation. A size "
        "of 50 means that 50 bases on either side of the variant are "
        "used to compute the G+C content ")

    parser.set_defaults(
        methods=[],
        input_vcf_file=None,
        input_bed_file=None,
        region=None,
        input_fasta_file=None,
        format_distributions=[],
        format_distribution_nbins=1000,
        gc_window_size=50,
        report_step=1000000,
    )

    (args, unknown) = E.start(parser,
                              argv,
                              add_output_options=True,
                              unknowns=True)

    if len(unknown) == 1:
        args.input_vcf_file = unknown[0]

    if args.input_vcf_file is None:
        raise ValueError("please supply a VCF file")

    if args.input_fasta_file is None:
        raise ValueError("please supply a FASTA file")

    if "format-distribution" in args.methods and not args.format_distributions:
        raise ValueError("please supply at least one FORMAT field (DP, GQ) "
                         "when --method=format-distribution has been selected")

    if not os.path.exists(args.input_vcf_file):
        raise OSError("input vcf file {} does not exist".format(
            args.input_vcf_file))

    if not os.path.exists(args.input_vcf_file + ".tbi") and not \
       os.path.exists(args.input_vcf_file + ".csi"):
        raise OSError("input vcf file {} needs to be indexed".format(
            args.input_vcf_file))

    if not os.path.exists(args.input_fasta_file):
        raise OSError("input fasta file {} does not exist".format(
            args.input_fasta_file))

    if not os.path.exists(args.input_fasta_file + ".fai"):
        raise OSError("input fasta file {} needs to be indexed".format(
            args.input_fasta_file))

    # update paths to absolute
    args.input_fasta_file = os.path.abspath(args.input_fasta_file)
    args.input_vcf_file = os.path.abspath(args.input_vcf_file)

    # catch issue with empty variant files
    try:
        vcf_in = pysam.VariantFile(args.input_vcf_file)
    except (OSError, ValueError):
        E.warn("could not open variant file - likely to be empty")
        E.stop()
        return 0

    fasta_in = pysam.FastaFile(args.input_fasta_file)

    if args.input_bed_file:
        if not os.path.exists(args.input_bed_file):
            raise OSError("input bed file {} does not exist".format(
                args.input_bed_file))
        bed_in = pysam.TabixFile(args.input_bed_file)
    else:
        bed_in = None

    vcf2stats_count(
        vcf_in, fasta_in, bed_in, args)

    E.stop()


if __name__ == "__main__":
    sys.exit(main())



"""
bcl2fastq.py - Wrapper that outputs fastq files from bcl files
==============================================================

:Tags: BCL FASTQ Conversion

Purpose
-------

Convert the raw data from an Illumina Sequencing Run to :term:`fastq` formatted files.

Usage
-----

Example::

   python bcl2fastq.py -p "--runfolder-dir <RunFolder>"

This command demultiplexes and converts BCL files in the given run folder directory.
All arguments for Illumina's bcl2fastq software must be given with the -p argument. 
Type::

   python bcl2fastq.py --help

for command line help.

Documentation
-------------

Converts BCL files in a given run folder to fastq.

Command line options
--------------------

``--arguments``
    Supply arguments to be passed to Illumia's bcl2fastq software.

``--output-dir``
    Required if using --fastqc.

``--fastqc``
    After converting BCL files, run all fastq files in FastQC.

``--fastqc-options``
    Supply arguments to be passed to FastQC.

``--bcl2fastq-help``
    Prints help for Illumina's bcl2fastq software.

"""

import cgatcore.experiment as E
import subprocess
import sys
import glob
import gzip


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    parser = E.ArgumentParser()

    parser.add_argument("-p", "--arguments", type=str, dest="arguments",   
                        default="",
                        help="Pass options and arguments to the executable. Please surround options in \"\"")

    parser.add_argument("-o", "--output-dir", type=str, dest="output",
                        default=".",
                        help="Output for the fastq files.")

    parser.add_argument("-f", "--fastqc", dest="fastqc",
                        action="store_true",
                        help="After demultiplexing open the fastq files in FastQC.")

    parser.add_argument("-F", "--fastqc-options", type=str, dest="fastqc_options",
                        default="",
                        help="Options for FastQC. Please surround options in \"\"")

    parser.add_argument("-H", "--bcl2fastq-help", dest="bcl2fastq_help", 
                        action="store_true",
                        help="Print help for Illumina's bcl2fastq conversion software")

    (args) = E.start(parser)

    if subprocess.run("which bcl2fastq", shell=True).returncode:
        raise ValueError("bcl2fastq cannot be found")

    if args.bcl2fastq_help:
        subprocess.run("bcl2fastq --help", shell=True)
        return
    else:
        subprocess.run(f"bcl2fastq {args.arguments} -o {args.output}", shell=True)

    for infile in glob.glob(f"{args.output}/**/*.fastq.gz", recursive=True):
        with gzip.GzipFile(f"{infile}", "r") as f:
            if sum(1 for char in f.read().decode('utf-8') if char == "\n") % 4 != 0:
                raise ValueError(f"{infile} is either corrupt or incomplete.")

    if args.fastqc:
        for infile in glob.glob(f"{args.output}/**/*.fastq.gz", recursive=True):
            subprocess.run(f"fastqc {infile} {args.fastqc_options}", shell=True)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''bam2UniquePairs.py - filter/report uniquely mapped read pairs from a (bwa!) bam-file
======================================================================================

:Tags: Genomics NGS

Purpose
-------

Utility script to report and/or filter out "uniquely mapped" properly
paired reads

Reports:

1. The percentage of properly mapped read pairs with at least one
   uniquely mapped (XT=U) read

2. The percentage of properly mapped read pairs with at least one best
   mapped (X0-1) read

3. The percentage of properly mapped read pairs with at least one
   uniquely or best mapped (X0-1) read

If outfile is specified, reads are emitted when they are properly
paired and the pair has at least one read that is either best or
uniquely mapped.

Duplication is ignored.

Only BWA is supported.

TODO: cache and emit reads rather than iterating over the samfile twice...

'''

import sys
import cgatcore.experiment as E
import pysam


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument("-f", "--bam-file", "--filename", dest="filename", type=str,
                        help="bamfile")

    parser.add_argument("-a", "--aligner", dest="aligner", type=str,
                        help="bamfile", default="bwa")

    parser.add_argument("-r", "--output-report", type=str, dest="report",
                        help="bamfile", default="")

    parser.add_argument("-o", "--output-filename-bam", "--outfile", dest="outfile", type=str,
                        help="bamfile", default="")

    # add common options (-h/--help, ...) and parse command line
    (args) = E.start(parser, argv=argv, add_output_options=True)

    # Check the aligner is supported
    if args.aligner != "bwa":
        raise ValueError(
            "Currently only bwa is supported as aligner specific flags are used")

    # Check that either a report or outfile name has been specified
    if args.report == "" and args.outfile == "":
        raise ValueError("Nothing to do")

    # Analyse the bamfile
    samfile = pysam.AlignmentFile(args.filename, "rb")
    uniq_map, best_map, uORb_map = {}, {}, {}
    properly_paired = 0

    for read in samfile.fetch():

        if read.is_proper_pair:
            tagd = dict(read.tags)
            u, b, key = False, False, read.qname

            if tagd["XT"] == "U":
                u = True
                uniq_map[key] = 1

            if "X0" in tagd:
                if tagd["X0"] == 1:
                    b = True
                    best_map[key] = 1

            if u is True or b is True:
                uORb_map[key] = 1

            properly_paired += 1

    samfile.close()

    npp = properly_paired / 2

    E.info("No proper pairs: %s" % npp)

    # Write a tabular report if report name given
    if args.report != "":

        E.info("Writing report on no. proper pairs with unique/best reads")

        def _row(x, npp=npp):
            name, d = x
            n = len(list(d.keys()))
            pc = float(n) / npp * 100
            line = "%s\t%i\t%.2f" % (name, n, pc)
            return(line)

        header = "\t".join(["pair_criteria", "n_proper_pairs",
                            "percent_proper_pairs"])

        with iotools.open_file(args.report, "w") as report:
            report.write(header + "\n")
            for x in [("unique", uniq_map), ("best", best_map),
                      ("unique_or_best", uORb_map)]:
                report.write(_row(x) + "\n")

    # Create new bam containing uniquely mapping read pairs
    # if outfile specified
    if args.outfile != "":

        E.info("Writing proper pairs with unique or best read to %s" %
               args.outfile)

        samfile = pysam.AlignmentFile(args.filename, "rb")
        outbam = pysam.AlignmentFile(args.outfile, "wb", template=samfile)

        for read in samfile.fetch():
            if read.is_proper_pair:
                if read.qname in uORb_map:
                    outbam.write(read)
        samfile.close()
        outbam.close()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
csv2csv.py - operate on tables
==============================

:Tags: Python

Purpose
-------

operate on tables.

Usage
-----

Example::

   python csv2csv.py --help

Type::

   python csv2csv.py --help

for command line help.

Command line options
--------------------

'''
import sys
import csv
import cgatcore.experiment as E
import cgatcore.iotools as iotools


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument(
        "-s", "--method=sort --sort-order", dest="sort", type=str,
        help="fields to take (in sorted order).")

    (args) = E.start(parser, add_csv_options=True)

    reader = csv.DictReader(E.stdin, dialect=args.csv_dialect)

    if args.sort:
        fields = args.sort.split(",")
    else:
        fields = None

    writer = csv.DictWriter(E.stdout,
                            fields,
                            dialect=args.csv_dialect,
                            lineterminator=args.csv_lineterminator,
                            extrasaction='ignore')

    E.stdout.write("\t".join(fields) + "\n")

    for row in reader:
        row = iotools.convertDictionary(row)
        writer.writerow(row)

    E.stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''csv_cut.py - select columns from a table
========================================

:Tags: Python

Purpose
-------

extract named columns from a csv formatted table


.. todo::

   describe purpose of the script.

Usage
-----

Extract the two columns gene and length from a table in standard input::

   python csv_cut.py gene length < stdin

The script permits the use of patterns. For example, the command will
select the column gene and all columns that contain the part 'len'::

   python csv_cut.py gene %len% < stdin

Type::

   python csv_cut.py --help

for command line help.

Command line options
--------------------

'''
import sys
import re
import cgatcore.experiment as E
import csv
import six
import _csv
import hashlib
from cgatcore.csvutils import CommentStripper, DictReaderLarge


class UniqueBuffer:
    mKeys = {}

    def __init__(self, outfile):
        self.mOutfile = outfile

    def write(self, out):
        key = hashlib.md5(out).digest()
        if key not in self.mKeys:
            self.mKeys[key] = True
            self.mOutfile.write(out)


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument("-r", "--remove", dest="remove", action="store_true",
                        help="remove specified columns, keep all others.")

    parser.add_argument("-u", "--unique", dest="unique", action="store_true",
                        help="output rows are uniq.")

    parser.add_argument("-l", "--large", dest="large", action="store_true",
                        help="large columns. Do not use native python csv module.")

    parser.add_argument("-f", "--filename-fields", dest="filename_fields", type=str,
                        help="filename with field information.")

    parser.set_defaults(
        remove=False,
        unique=False,
        large=False,
        filename_fields=None,
    )

    (args, unknown) = E.start(parser,
                              add_csv_options=True,
                              quiet=True,
                              unknowns=True)

    input_fields = unknown

    if args.filename_fields:
        input_fields = [x[:-1].split("\t")[0] for x in [x for x in iotools.open_file(args.filename_fields, "r").readlines() if x[0] != "#"]]

    if args.unique:
        outfile = UniqueBuffer(args.stdout)
    else:
        outfile = args.stdout

    while 1:
        line = args.stdin.readline()

        if not line:
            E.stop()
            sys.exit(0)

        if line[0] == "#":
            continue

        first_line = line
        break

    old_fields = first_line[:-1].split("\t")

    fields = []
    for f in input_fields:
        # do pattern search
        if f[0] == "%" and f[-1] == "%":
            pattern = re.compile(f[1:-1])
            for o in old_fields:
                if pattern.search(o) and o not in fields:
                    fields.append(o)
        else:
            if f in old_fields:
                fields.append(f)

    if args.remove:
        fields = set(fields)
        fields = [x for x in old_fields if x not in fields]

    if args.large:
        reader = DictReaderLarge(CommentStripper(args.stdin),
                                 fieldnames=old_fields,
                                 dialect=args.csv_dialect)
    else:
        reader = csv.DictReader(CommentStripper(args.stdin),
                                fieldnames=old_fields,
                                dialect=args.csv_dialect)

    writer = csv.DictWriter(outfile,
                            fields,
                            dialect=args.csv_dialect,
                            lineterminator=args.csv_lineterminator,
                            extrasaction='ignore')

    print("\t".join(fields))

    first_row = True
    ninput, noutput, nerrors = 0, 0, 0

    while 1:
        ninput += 1
        try:
            row = six.next(reader)
        except _csv.Error as msg:
            args.stderr.write("# error while parsing: %s\n" % (msg))
            nerrors += 1
            continue
        except StopIteration:
            break
        if not row:
            break
        writer.writerow(row)
        noutput += 1

    E.info("ninput=%i, noutput=%i, nerrors=%i" % (ninput, noutput, nerrors))

    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
csv_rename.py - rename columns in a table
=========================================

:Tags: Python

Purpose
-------

rename columns in a csv file

Usage
-----

Example::

   csv_rename.py gene=id < stdin

Type::

   python csv_rename.py --help

for command line help.

Command line options
--------------------

'''
import sys
import cgatcore.experiment as E


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument("-r", "--remove", dest="remove", action="store_true",
                        help="remove specified columns, keep all others.")

    parser.add_argument("-u", "--unique", dest="unique", action="store_true",
                        help="output rows are uniq.")

    parser.add_argument("-f", "--filename-fields", dest="filename_fields", type=str,
                        help="filename with field information.")

    parser.set_defaults(
        filename_fields=None,
    )

    (args) = E.start(parser,
                     add_csv_options=True)
    mapper = {}
    for x in args:
        a, b = x.split("=")
        mapper[a.strip()] = b.strip()

    while 1:
        line = args.stdin.readline()

        if not line:
            E.stop()
            sys.exit(0)

        if line[0] == "#":
            args.stdout.write(line)
            continue

        break

    header = []
    nreplaced = 0
    for x in line[:-1].split():
        if x in mapper:
            nreplaced += 1
            header.append(mapper[x])
        else:
            header.append(x)

    args.stdout.write("\t".join(header) + "\n")
    nlines = 0
    for line in args.stdin:
        nlines += 1
        args.stdout.write(line)

    if args.loglevel >= 1:
        ninput = len(header)
        noutput = ninput
        args.stdout.write("# ninput=%i, noutput=%i, nreplaced=%i, nlines=%i\n" % (
            ninput, noutput, nreplaced, nlines))

    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""
index2bed.py - convert indexed fasta file to bed file
=====================================================

:Tags: Python

Purpose
-------

Usage
-----

Type::

   python <script_name>.py --help

for command line help.

Command line options
--------------------

"""

import sys
import re
import cgat.IndexedFasta as IndexedFasta
import cgatcore.experiment as E


def getFixedWidthWindows(map_contig2size, window_size):
    """return a list of fixed contig sizes."""

    for contig, size in list(map_contig2size.items()):
        E.info("processing %s" % contig)
        for x in range(0, size, window_increment):
            if x + window_size > size:
                continue
            gff = GTF.Entry()
            gff.feature = "window"
            gff.source = "window"
            gff.contig = contig
            gff.start = x
            gff.end = min(size, x + window_size)
            yield gff


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument(
        "-g", "--genome-file", dest="genome_file", type=str,
        help="filename with genome.")

    parser.add_argument(
        "--remove-regex", dest="remove_regex",
        type=str,
        help="regular expression of contigs to remove.")

    parser.add_argument(
        "-e", "--gff-file", dest="gff_file", type=str,
        help="gff file to use for getting contig sizes.")

    parser.add_argument(
        "-f", "--fixed-width-windows",
        dest="fixed_width_windows", type=str,
        help="fixed width windows. Supply the window size as a "
        "parameter. Optionally supply an offset.")

    parser.set_defaults(
        genome_file=None,
        remove_regex=None,
        fixed_windows=None,
    )

    # add common options (-h/--help, ...) and parse command line
    (args) = E.start(parser, argv=argv)

    if args.remove_regex:
        remove_regex = re.compile(args.remove_regex)
    else:
        remove_regex = None

    if args.fixed_width_windows:
        v = list(map(int, args.fixed_width_windows.split(",")))
        if len(v) == 2:
            window_size, window_increment = v
        elif len(v) == 1:
            window_size, window_increment = v[0], v[0]
        else:
            raise ValueError(
                "could not parse window size '%s': should be size[,increment]" % args.fixed_width_windows)

    if args.gff_file:
        infile = iotools.open_file(args.gff_file, "r")
        gff = GTF.readFromFile(infile)
        infile.close()
        for g in gff:
            try:
                map_contig2size[g.mName] = max(map_contig2size[g.mName], g.end)
            except ValueError:
                map_contig2size[g.mName] = g.end

    else:
        gff = None

    if args.genome_file:
        fasta = IndexedFasta.IndexedFasta(args.genome_file)
        map_contig2size = fasta.getContigSizes(with_synonyms=False)
    else:
        fasta = None

    if map_contig2size is None:
        raise ValueError("no source of contig sizes supplied")

    # do sth
    counter = E.Counter()

    for contig, size in list(map_contig2size.items()):
        size = int(size)
        counter.input += 1

        if remove_regex and remove_regex.search(contig):
            counter.skipped += 1
            continue

        if args.fixed_width_windows:
            for x in range(0, size, window_increment):
                if x + window_size > size:
                    continue
                args.stdout.write(
                    "%s\t%i\t%i\n" % (contig, x, min(size, x + window_size)))
                counter.windows += 1
        else:
            args.stdout.write("%s\t%i\t%i\n" % (contig, 0, size))
            counter.windows += 1

        counter.output += 1

    E.info(str(counter))

    # write footer and output benchmark information.
    E.stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
cgat_get_options.py - build a sorted list of all options used in scripts
========================================================================

:Author:
:Tags: Python

Purpose
-------

Go through all scripts in the cgat code collection and collect
options used in the scripts.

This script expects to be executed at the root of the
cgat code repository.


Usage
-----

.. Example use case

Example::

   python cgat_get_options.py

Type::

   python cgat_get_options.py --help

for command line help.

Command line options
--------------------

'''

import sys
import os
import glob
import imp
import collections
import pandas
import cgatcore.experiment as E
import cgatcore.iotools as iotools

ORIGINAL_START = None

PARSER = None

EXPRESSIONS = (
    ('scripts', 'scripts/*.py'),)

EXCLUDE = ("__init__.py",
           "cgat.py",)


class DummyError(Exception):
    pass


def LocalStart(parser, *args, **kwargs):
    '''stub for E.start - set return_parser argument to true'''
    global PARSER
    PARSER = ORIGINAL_START(parser,
                            return_parser=True,
                            **kwargs
                            )
    raise DummyError()


def collectOptionsFromScript(script_name):
    '''collect options used in script *script_name*.'''

    # call other script
    prefix, suffix = os.path.splitext(script_name)

    dirname = os.path.dirname(script_name)
    basename = os.path.basename(script_name)[:-3]

    if os.path.exists(prefix + ".pyc"):
        os.remove(prefix + ".pyc")

    # check if script contains getopt
    with iotools.open_file(script_name) as inf:
        if "getopt" in inf.read():
            E.warn("script %s uses getopt directly" % script_name)
            return []

    try:
        module = imp.load_source(basename, script_name)
    except ImportError as msg:
        E.warn('could not import %s - skipped: %s' % (basename, msg))
        return []

    E.start = LocalStart

    try:
        module.main(argv=["--help"])
    except AttributeError:
        E.warn("no main method in %s" % script_name)
        return []
    except SystemExit:
        E.warn("script exits - possibly does not use E.start()")
        return []
    except DummyError:
        pass

    result = []
    for option in PARSER.option_list:
        # ignore options added by optparse
        if option.dest is None:
            continue

        optstring = option.get_opt_string()
        if optstring.startswith("--"):
            optstring = optstring[2:]
        result.append(optstring)

    return result


def main(argv=None):
    """script main.
    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument(
        "--inplace", dest="inplace", action="store_true",
        help="update option list in place. New options will"
        "be added to the list given by --options-tsv-file. "
        "Options will only be added, not removed ")

    parser.add_argument(
        "--options-tsv-file", dest="tsv_file", type=str,
        help="existing table with options. Will be updated if "
        "--in-place is set [default]")

    parser.set_defaults(
        inplace=False,
        tsv_file=None)

    # add common options (-h/--help, ...) and parse command line
    (args) = E.start(parser, argv=argv)

    old_options = None
    if args.tsv_file:
        if not os.path.exists(args.tsv_file):
            raise OSError(
                "filename %s not found, see --options-tsv-file" %
                args.tsv_file)
        old_options = pandas.read_csv(
            iotools.open_file(args.tsv_file),
            sep="\t",
            index_col=0,
        )
        old_options = old_options.fillna("")

    global ORIGINAL_START
    ORIGINAL_START = E.start

    all_options = collections.defaultdict(list)

    for label, expression in EXPRESSIONS:

        files = glob.glob(expression)
        files.sort()

        for f in files:

            E.debug("processing %s" % f)
            if os.path.isdir(f):
                continue
            if os.path.basename(f) in EXCLUDE:
                continue
            collected_options = collectOptionsFromScript(os.path.abspath(f))
            for o in collected_options:
                all_options[o].append(f)

    # add old options
    for x in old_options.index:
        if x not in all_options:
            all_options[x].append("--")

    if args.inplace:
        outfile = iotools.open_file(args.tsv_file, "w")
        E.info("updating file '%s'" % args.tsv_file)
    else:
        outfile = args.stdout

    outfile.write("option\taction\tcomment\talternative\tfiles\n")
    for o, v in sorted(all_options.items()):
        try:
            action, comment, alternative, ff = old_options.xs(o)

        except KeyError:
            action, comment, alternative, ff = "", "", "", ""

        if comment == "nan":
            comment = ""
        if alternative == "nan":
            alternative = ""

        outfile.write("\t".join((list(map(
            str, (o, action, comment, alternative, ",".join(v)))))) + "\n")

    if outfile != args.stdout:
        outfile.close()

    # write footer and output benchmark information.
    E.stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''wig2bed.py - convert densities to intervals
===========================================

Purpose
-------

define intervals based on densities within a bigwig file.

The script currently implements the following methods (``--method``):

threshold
    output windows that contain values above a certain
    threshold.

std-above-mean
    output windows that are a certain number of standard
    deviations above the mean.

multiple-of-mean
    output windows that are a certain times above the mean.

Usage
-----

Bigwig files need to be supplied by the --bigwig-file options.

For example::

    python wig2bed.py --threshold=10 --method=threshold --genome-file=mm10 --bigwig-file=in.bw > out.bed

Command line options
--------------------

'''

import sys
import re
import collections

import cgatcore.experiment as E
import cgat.IndexedFasta as IndexedFasta
import cgatcore.iotools as iotools

import pyBigWig


def block_iterator(infile, contig, size, chunk_size=10000000):

    for x in range(0, size, chunk_size):
        iterator = infile.get(contig, x, x + chunk_size)
        if iterator is None:
            raise StopIteration
        for v in iterator:
            yield v


def applyThreshold(infile, fasta, threshold, max_distance=0):
    '''apply threshold to a wig file writing a
    bed-formatted file as output.'''

    c = E.Counter()

    for contig, size in list(fasta.getContigSizes(with_synonyms=False).items()):
        c.contigs += 1

        E.debug("processing %s" % contig)

        last_start, last_end = -1, 0

        for start, end, value in block_iterator(infile, contig, size):
            d = start - last_end
            if (d > 0 or value < threshold):
                if last_start >= 0:
                    yield contig, last_start, last_end
                    c.intervals += 1
                last_start = -1
            elif last_start < 0 and value >= threshold:
                last_start = start

            last_end = end

        if last_start >= 0:
            yield contig, last_start, end
            c.intervals += 1

        c.output += 1

    E.info(str(c))


def getBigwigSummary(bigwig_file):
    '''return summary of bigwig contents.

    This method uses the bigWigInfo UCSC utility
    '''

    results = E.run("bigWigInfo %(bigwig_file)s" %
                    locals(), return_stdout=True)

    data = [x.split(":") for x in results.split("\n") if x != ""]
    fields = [x[0] for x in data]
    Results = collections.namedtuple("BigwigInfo", fields)

    def conv(v):
        return iotools.str2val(re.sub(",", "", v.strip()))

    results = Results(*[conv(x[1]) for x in data])
    return results


def main(argv=sys.argv):

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument("-m", "--method", dest="methods", type=str,
                        action="append",
                        choices=("threshold", "stddev-above-mean",
                                 "multiple-of-mean"),
                        help="method to apply")

    parser.add_argument("-g", "--genome-file", dest="genome_file", type=str,
                        help="filename with genome.")

    parser.add_argument("-t", "--threshold", dest="threshold", type=float,
                        help="threshold to apply")

    parser.add_argument(
        "-i", "--bigwig-file", dest="bigwig_file",
        type=str, metavar="bigwig",
        help="filename with bigwig information.")

    parser.set_defaults(methods=[],
                        genome_file=None,
                        threshold=10,
                        max_distance=0)

    (args) = E.start(parser, add_pipe_options=True)

    if args.bigwig_file:
        bigwig_file = pyBigWig.open(args.bigwig_file)
    else:
        bigwig_file = None

    if args.genome_file:
        genome_fasta = IndexedFasta.IndexedFasta(args.genome_file)
        contigs = genome_fasta.getContigSizes()

    for method in args.methods:
        if method == "threshold":
            if not contigs:
                raise ValueError("please supply contig sizes")
            if not bigwig_file:
                raise NotImplementedError(
                    "threshold not implemented for wig files")
            processor = applyThreshold(bigwig_file,
                                       genome_fasta,
                                       threshold=args.threshold,
                                       max_distance=args.max_distance)
        elif method == "stddev-above-mean":
            if not contigs:
                raise ValueError("please supply contig sizes")
            if not bigwig_file:
                raise NotImplementedError(
                    "threshold not implemented for wig files")
            summary = getBigwigSummary(args.bigwig_file)
            threshold = summary.mean + args.threshold * summary.std
            E.info("applying threshold %f: mean=%f, std=%f" %
                   (threshold, summary.mean, summary.std))
            processor = applyThreshold(bigwig_file,
                                       genome_fasta,
                                       threshold=threshold,
                                       max_distance=args.max_distance)

        elif method == "multiple-of-mean":
            if not contigs:
                raise ValueError("please supply contig sizes")
            if not bigwig_file:
                raise NotImplementedError(
                    "threshold not implemented for wig files")
            summary = getBigwigSummary(args.bigwig_file)
            threshold = summary.mean * args.threshold
            E.info("applying threshold %f: mean=%f, std=%f" %
                   (threshold, summary.mean, summary.std))
            processor = applyThreshold(bigwig_file,
                                       genome_fasta,
                                       threshold=threshold,
                                       max_distance=args.max_distance)

    outfile = args.stdout

    outfile.write("".join(["%s\t%i\t%i\n" % x for x in processor]))
    outfile.write("\n")

    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
bed.plot.py - create genomic snapshots using the IGV Viewer
===========================================================

:Tags: Python

Purpose
-------

Create genomic plots in a set of intervals using
the IGV snapshot mechanism.

The script can use a running instance of IGV identified
by host and port. Alternatively, it can start IGV and load
a pre-built session.

Usage
-----

Example::

   python bed2plot.py < in.bed

Type::

   python script_template.py --help

for command line help.

Command line options
--------------------

'''

import os
import sys
import re
import socket
import pysam


import cgatcore.experiment as E


class IGV(object):
    """based on IGV.py by Brent Petersen, see here:
    https://github.com/brentp/bio-playground/blob/master/igv/igv.py

    (MIT licenced)
    """

    _socket = None
    _path = None

    def __init__(self, host='127.0.0.1', port=60151, snapshot_dir='/tmp/igv'):
        self.host = host
        self.port = port
        self.commands = []
        self.connect()
        self.set_path(snapshot_dir)

    def connect(self):
        if self._socket:
            self._socket.close()
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.connect((self.host, self.port))

    def go(self, position):
        return self.send('goto ' + position)
    goto = go

    def genome(self, name):
        return self.send('genome ' + name)

    def load(self, url):
        return self.send('load ' + url)

    def region(self, contig, start, end):
        return self.send(' '.join(map(str, ['region', contig, start, end])))

    def sort(self, option='base'):
        """
        options is one of: base, position, strand, quality, sample, and
        readGroup.
        """
        assert option in ("base", "position", "strand", "quality", "sample",
                          "readGroup")
        return self.send('sort ' + option)

    def set_path(self, snapshot_dir):
        if snapshot_dir == self._path:
            return
        if not os.path.exists(snapshot_dir):
            os.makedirs(snapshot_dir)

        self.send('snapshotDirectory %s' % snapshot_dir)
        self._path = snapshot_dir

    def expand(self, track=''):
        self.send('expand %s' % track)

    def collapse(self, track=''):
        self.send('collapse %s' % track)

    def clear(self):
        self.send('clear')

    def send(self, cmd):
        # socket in Python2 oprates with strings
        if sys.version_info.major == 2:
            self._socket.send(cmd + '\n')
            return self._socket.recv(4096).rstrip('\n')
        # while socket in Python3 requires bytes
        else:
            self.commands.append(cmd)
            cmd = cmd + '\n'
            self._socket.send(cmd.encode('utf-8'))
            return self._socket.recv(4096).decode('utf-8').rstrip('\n')

    def save(self, path=None):
        if path is not None:
            # igv assumes the path is just a single filename, but
            # we can set the snapshot dir. then just use the filename.
            dirname = os.path.dirname(path)
            if dirname:
                self.set_path(dirname)
            return self.send('snapshot ' + os.path.basename(path))
        else:
            return self.send('snapshot')
    snapshot = save


def main(argv=sys.argv):

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("-s", "--session", dest="session",
                        type=str,
                        help="load session before creating plots "
                        )

    parser.add_argument("-d", "--snapshot-dir", dest="snapshotdir",
                        type=str,
                        help="directory to save snapshots in ")

    parser.add_argument("-f", "--format", dest="format", type=str,
                        choices=("png", "eps", "svg"),
                        help="output file format ")

    parser.add_argument("-o", "--host", dest="host", type=str,
                        help="host that IGV is running on ")

    parser.add_argument("-p", "--port", dest="port", type=int,
                        help="port that IGV listens at ")

    parser.add_argument("-e", "--extend", dest="extend", type=int,
                        help="extend each interval by a number of bases "
                        )

    parser.add_argument("-x", "--expand", dest="expand", type=float,
                        help="expand each region by a certain factor "
                        )

    parser.add_argument("--session-only", dest="session_only",
                        action="store_true",
                        help="plot session after opening, "
                        "ignore intervals "
                        )

    parser.add_argument("-n", "--name", dest="name", type=str,
                        choices=("bed-name", "increment"),
                        help="name to use for snapshot "
                        )

    parser.set_defaults(
        command="igv.sh",
        host='127.0.0.1',
        port=61111,
        snapshotdir=os.getcwd(),
        extend=0,
        format="png",
        expand=1.0,
        session=None,
        session_only=False,
        keep_open=False,
        name="bed-name",
    )

    # add common options (-h/--help, ...) and parse command line
    (args) = E.start(parser, argv=argv, add_output_options=True)

    igv_process = None
    if args.new_instance:
        E.info("starting new IGV process")
        igv_process = IGV.startIGV(command=args.command,
                                   port=args.port)
        E.info("new IGV process started")

    E.info("connection to process on %s:%s" % (args.host, args.port))
    E.info("saving images in %s" % args.snapshotdir)
    igv = IGV(host=args.host,
              port=args.port,
              snapshot_dir=os.path.abspath(args.snapshotdir))

    if args.session:
        E.info('loading session from %s' % args.session)
        igv.load(args.session)
        E.info('loaded session')

    if args.session_only:
        E.info('plotting session only ignoring any intervals')
        fn = "%s.%s" % (os.path.basename(args.session), args.format)
        E.info("writing snapshot to '%s'" %
               os.path.join(args.snapshotdir, fn))
        igv.save(fn)

    else:
        c = E.Counter()
        for bed in pysam.tabix_iterator(args.stdin,
                                        parser=pysam.asBed()):

            c.input += 1

            # IGV can not deal with white-space in filenames
            if args.name == "bed-name":
                name = re.sub("\s", "_", bed.name)
            elif args.name == "increment":
                name = str(c.input)

            E.info("going to %s:%i-%i for %s" %
                   (bed.contig, bed.start, bed.end, name))

            start, end = bed.start, bed.end
            extend = args.extend
            if args.expand:
                d = end - start
                extend = max(extend, (args.expand * d - d) // 2)

            start -= extend
            end += extend

            igv.go("%s:%i-%i" % (bed.contig, start, end))

            fn = E.get_output_file("%s.%s" % (name, args.format))
            E.info("writing snapshot to '%s'" % fn)
            igv.save(fn)

            c.snapshots += 1

        E.info(c)

    if igv_process is not None and not args.keep_open:
        E.info('shutting down IGV')
        igv_process.send_signal(signal.SIGKILL)

    E.stop()


if __name__ == "__main__":
    sys.exit(main())



'''fastqs2fastq.py - merge reads in fastq files
============================================

:Tags: Genomics NGS FASTQ FASTQ Manipulation

Purpose
-------

This script takes two paired-ended fastq files and outputs a single
fastq file in which reads have merged.

The two files must be sorted by read identifier.

Note that this script is currently a proof-of-principle implementation
and has not been optimized for speed or functionality.

Usage
-----

Example::

   python fastqs2fastq.py myReads.1.fastq.gz myReads.2.fastq.gz
          --method=join
          > join.fastq

In this example we take a pair of fastq files, join the reads and save
the output in :file:`join.fastq`.

Type::

   python fastqs2fastq.py --help

for command line help.

Command line options
--------------------

'''

import sys
import collections
import copy

import cgatcore.iotools as iotools
import cgatcore.experiment as E
import cgat.Fastq as Fastq
import cgat.Genomics as Genomics


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("-m", "--method", dest="method", type=str,
                        choices=('join', ),
                        help="method to apply.")

    parser.set_defaults(
        method="join",
    )

    # add common options (-h/--help, ...) and parse command line
    (args, unknown) = E.start(parser,
                              argv=argv,
                              unknowns=True)

    if len(unknown) != 2:
        raise ValueError(
            "please supply at least two fastq files on the commandline")

    fn1, fn2 = unknown
    c = E.Counter()
    outfile = args.stdout

    if args.method == "join":
        # merge based on diagonals in dotplot
        iter1 = Fastq.iterate(iotools.open_file(fn1))
        iter2 = Fastq.iterate(iotools.open_file(fn2))
        tuple_size = 2
        for left, right in zip(iter1, iter2):
            c.input += 1

            # build dictionary of tuples
            s1, q1 = left.seq, left.quals
            d = collections.defaultdict(list)
            for x in range(len(s1) - tuple_size):
                d[s1[x:x + tuple_size]].append(x)

            s2, q2 = right.seq, right.quals
            s2 = Genomics.reverse_complement(s2)
            q2 = q2[::-1]

            # compute list of offsets/diagonals
            offsets = collections.defaultdict(int)
            for x in range(len(s2) - tuple_size):
                c = s2[x:x + tuple_size]
                for y in d[c]:
                    offsets[x - y] += 1

            # find maximum diagonal
            sorted = sorted([(y, x) for x, y in list(offsets.items())])
            max_count, max_offset = sorted[-1]

            E.debug('%s: maximum offset at %i' % (left.identifier,
                                                  max_offset))

            # simple merge sequence
            take = len(s2) - max_offset
            merged_seq = s1 + s2[take:]

            # simple merge quality scores
            merged_quals = q1 + q2[take:]

            new_entry = copy.copy(left)
            new_entry.seq = merged_seq
            new_entry.quals = merged_quals
            outfile.write(new_entry)
            c.output += 1

    # write footer and output benchmark information.
    E.info("%s" % str(c))
    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''table2table.py - operate on tables
==================================

:Tags: Python

Purpose
-------

This script implements a few methods for manipulating tables.

Methods working on all tables:
++++++++++++++++++++++++++++++

transpose
   transpose a table

split-fields
   Split muliple-value fields in each row at ``--separator``. Output
   multiple rows with all combinations.

group
   Group values by column

join-column
   Join rows in a table by columns

expand-table
   If a field in a row contains multiple values,
   the row is expanded into multiple rows such
   that all values have space.

flatten-table
   Output a table as row/column/value tuples.

as-column
   Output table as a single column. Colums in the original table are
   appended and output.

collapse-table
   Collapse a table of two columns with row names in the first
   column. Outputs a table with multiple columns for each row name.

Methods for numerical columns
+++++++++++++++++++++++++++++

Some methods make only sense for columns containing numerical values.
If a table contains both numerical and non-numerical data, the
numerical columns can be specified by the ``--columns`` option.

normalize-by-value
   divide all cells in a table by a value

multiply-by-value
   multiply all cells in a table by a value

lower-bound
   replace all cells with a value of less than lower bound with the lower
   bound.

upper-bound
   replace all cells with a value of more than upper bound with the upper
   bound.

normalize-by-table
   divide each cell in a table with the corresponding entry in a secondary
   table.

normalize-by-max
   divide table columns by maximum per column

kullback-leibler
   compute kullback-leibler divergence between two columns. Compute
   both D(a||b), D(b||a) and (D(a||b) + D(b||a)) / 2

rank
   substitute cells with their ranks in a column

fdr
   compute an FDR over selected columns. Replaces the columns
   with the qvalues.

Usage
-----

Example::

   python table2table.py --help

Type::

   python table2table.py --help

for command line help.

Command line options
--------------------

'''
import sys
import cgatcore.table
from cgatcore.table import main
from cgat import Stats

# monkeypatch: import Stats in cgatcore.Table's namespace.
cgatcore.table.Stats = Stats


if __name__ == "__main__":
    sys.exit(main())



'''gff32gtf.py - various methods for converting gff3 files to gtf
====================================================

:Tags: Python

Purpose
-------

Provide a range of methods for converting GFF3 formated files to valid GTF 
format files.

Background
----------

While the various flavours of GFF format are supposedly backward
compatible, this is broken by GTF2.2 and GFF3. GTF requires the
presence of gene_id and transcript_id fields for each record. This not
so for GFF3. Further key,value tags in the attributes fields of GTF
are " " delimited, but are "=" delimited in GFF.

Conversion is non-trivial. GFF3 records are hierachical. To find the
gene_id and transcript_id one must traverse the hierarchy to the
correct point. Futher records can have multiple parents.

                                               -> Exon
While the standard structure is Gene -> mRNA -|       ,
                                               -> CDS

this is not manditory, and it is possible the conversion will want to
be done in a different way.

Usage
-----

Example::

   python gff32gtf.py --method=[METHOD] [options]

Their are several ways in which the conversion can be done:

hierachical
+++++++++++

By default this script will read in the entire GFF3 file, and then for
each entry traverse the hierarchy until an object of type GENE_TYPE
("gene" by default") or an object with no parent is found. This
becomes the "gene_id". Any object of TRANSCRIPT_TYPE encountered on
the way is set as the transcript_id. If not such object is encountered
then the object directly below the gene object is used as the
trancript_id. Objects that belong to multipe transcripts or genes are
duplicated.

This method requires ID and Parent fields to be present.

Because this method reads the whole file in, it uses the most memory, although
see --read-twice and --by-chrom for tricks that might help.

set-field
+++++++++

The gene_id and transcript_id fields are set to the  value of a provided field.
Records that don't have these fields are discarded. By default:

transcript_id=ID
gene_id=Parent

set-pattern
+++++++++++

As above, but the fieldnames are set by a string format involving the
fields of the record.

set-none
++++++++

transcript_id and gene_id are set to None.

Command line options
--------------------

'''

import sys

import cgatcore.experiment as E
import cgat.GFF3 as GFF3
import cgat.GTF as GTF
import cgatcore.iotools as iotools


def search_hierarchy(ID, hierarchy, options):
    '''Returns a three element tuple of lists.

        * The first two lists are the gene_ids and transcript_ids that
        * are associated with specified IDs.  The third is a list of
        * possible transcript_ids - that is trancript_ids that are one
        * level below where the gene id came from.

        All three lists are guarenteed to be the same length, but both
        the transcript lists could contain None values where no
        transcript_id has been found.

        Works by calling it self recursively, not efficient, but does
        deal with the problem of cicular references: the recursion
        limit will be quickly reached.

        Can also raise ValueError if no feature of type
        options.gene_type is found and options.missing_gene is false

    '''

    gene_id = []
    transcript_id = []
    possible_transcript_id = []

    entry = hierarchy[ID]

    if entry['type'] == options.gene_type:
        gene_id.append(hierarchy[ID]['gene_id'])

        if not entry['type'] == options.transcript_type:
            transcript_id = [None]
            possible_transcript_id = [None]
        else:
            transcript_id = [entry['transcript_id']]
            possible_transcript_id = [None]

        return (gene_id, transcript_id, possible_transcript_id)

    for parent in entry['Parent']:

        new_gene_id, new_transcript_id, new_possible_transcript_id = search_hierarchy(
            parent, hierarchy, options)

        gene_id.extend(new_gene_id)
        transcript_id.extend(new_transcript_id)
        possible_transcript_id.extend(new_possible_transcript_id)

    if options.missing_gene:
        possible_transcript_id = [
            entry['transcript_id'] if x is None else x for x in possible_transcript_id]

    if len(gene_id) == 0 and options.missing_gene:
        gene_id = [entry['gene_id']]
        transcript_id = [None]
        possible_transcript_id = [None]
    elif len(gene_id) == 0 and not options.missing_gene:
        raise ValueError(
            "Root found without finding an object of type %s" % options.gene_type)

    if entry['type'] == options.transcript_type:
        transcript_id = [
            entry['transcript_id'] if x is None else x for x in transcript_id]

    assert len(gene_id) == len(transcript_id) and len(
        transcript_id) == len(possible_transcript_id)
    assert len(gene_id) > 0

    return gene_id, transcript_id, possible_transcript_id


def convert_hierarchy(first_gffs, second_gffs, options):
    ''' Converts GFF to GTF by parsing the hierarchy.
    First parses :param:first_gffs to build the hierarchy then iterates over second_gffs
    using a call to the recursive function search_hierarchy to identify gene_ids and transcript_ids.

    If multiple gene and transcript_ids are found outputs a record for each combination.

    If no definitive transcript_id is found and options.missing_gene is True, it will use the
    possible_transcript_id as transcript_id, which is the ID one level below the entry used as gene_id.
    If this is also None (that is there was only on level), sets transcript_id to gene_id.

    Might raise ValueError if options.missing_gene is false and either no gene or no transcript_id
    was found for an entry.

    Might raise RuntimeError if the recursion limit was reached because the input contains circular
    references. '''

    hierarchy = {}

    for gff in first_gffs:

        if not(options.parent == "Parent"):
            if options.parent in gff.asDict():
                gff['Parent'] = gff[options.parent].split(",")
            else:
                gff['Parent'] = []

        hierarchy[gff['ID']] = {
            "type": gff.feature,
            "Parent": gff.asDict().get("Parent", []),
            "gene_id": gff.attributes.get(
                options.gene_field_or_pattern, gff['ID']),
            "transcript_id": gff.attributes.get(
                options.transcript_field_or_pattern, gff['ID'])}

    for gff in second_gffs:

        if options.discard and (
                (options.missing_gene and options.parent not in gff) or (
                gff.feature in (options.gene_type, options.transcript_type))):

            continue

        gene_ids, transcript_ids, poss_transcript_ids = search_hierarchy(
            gff['ID'], hierarchy, options)

        assert len(gene_ids) > 0 and len(transcript_ids) > 0

        if options.missing_gene:

            transcript_ids = [poss if found is None else found
                              for found, poss in
                              zip(transcript_ids, poss_transcript_ids)]

            transcript_ids = [gid if found is None else found
                              for found, gid in
                              zip(transcript_ids, gene_ids)]

        elif None in transcript_ids:
            raise ValueError("failed to find transcript id for %s" % gff['ID'])

        for gene_id, transcript_id in zip(gene_ids, transcript_ids):

            gff.gene_id = gene_id
            gff.transcript_id = transcript_id

            gtf_entry = GTF.Entry()
            gtf_entry.copy(gff)
            if "Parent" in gtf_entry:
                gtf_entry['Parent'] = ",".join(gtf_entry['Parent'])

            options.stdout.write(str(gtf_entry) + "\n")


def convert_set(gffs, gene_pattern, transcript_pattern, options):
    ''' creates the gene_id and transcript_id fields from a string format pattern using
    fields of the gff. '''

    for gff in gffs:

        gff.gene_id = str(gene_pattern) % gff.asDict()
        gff.transcript_id = str(gene_pattern) % gff.asDict()

        gtf_entry = GTF.Entry()

        gtf_entry.copy(gff)
        if "Parent" in gtf_entry:
            gtf_entry['Parent'] = ",".join(gtf_entry['Parent'])

        options.stdout.write(str(gtf_entry) + "\n")


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument("-m", "--method", dest="method", type=str, action="store",
                        choices=("hierarchy", "set-field", "set-pattern", "set-none"),
                        help="Method to use for conversion")

    parser.add_argument("-g", "--gene-type", dest="gene_type", type=str,
                        help="feature type to get gene_id from if possible ")

    parser.add_argument("-t", "--transcript-type", dest="transcript_type", type=str,
                        help="feature type to get transcript_id from if possible ")

    parser.add_argument("-d", "--no-discard", dest="discard", action="store_false",
                        help="Do not discard feature types specified by GENE_TYPE and TRANSCRIPT_TYPE")

    parser.add_argument("--gene-id", dest="gene_field_or_pattern", type=str,
                        help="Either field or pattern for the gene_id ")

    parser.add_argument("--transcript-id", dest="transcript_field_or_pattern", type=str,
                        help="Either field or pattern for the transcript_id ")

    parser.add_argument("--parent-field", dest="parent", type=str,
                        help="field that specifies the parent relationship. Currently only"
                        "if left as Parent will features with multiple parents be parsed"
                        "correctly""")

    parser.add_argument("--read-twice", dest="read_twice", action="store_true",
                        help="Instead of holding the whole file in memory, read once for parsing the "
                        "hierarchy, and then again for actaully doing the conversion. Means a real file "
                        "and not a pipe must be provided.""")

    parser.add_argument("--by-chrom", dest="by_chrom", action="store_true",
                        help="Parse input file one choromosome at a time. Reduces memory usage, "
                        "but input must be sorted by chromosome and features may not split accross "
                        " multiple chromosomes""")

    parser.add_argument("--fail-missing-gene", dest="missing_gene", action="store_false",
                        help="Fail if no feature of type GENE_TYPE is found instead of using "
                        "defaulting to highest object in hierarchy""")

    parser.set_defaults(
        method="hierarchy",
        gene_type="gene",
        transcript_type="mRNA",
        discard=True,
        gene_field_or_pattern="ID",
        transcript_field_or_pattern="ID",
        read_twice=False,
        by_chrom=False,
        missing_gene=True,
        parent="Parent"
    )

    # add common options (-h/--help, ...) and parse command line
    (args) = E.start(parser, argv=argv)

    gffs = GFF3.flat_file_iterator(args.stdin)

    if args.by_chrom:
        gffs = GFF3.chrom_iterator(gffs)
    else:
        gffs = [gffs]

    # running early so that fails early if configuration is wrong
    if args.read_twice:
        # Will throw IOError if args.stdin is not a normal file
        second_gff = GFF3.flat_file_iterator(
            iotools.open_file(args.stdin.name))

        if args.by_chrom:
            second_gff = GFF3.chrom_iterator(second_gff)
        else:
            second_gff = iter([second_gff])
    else:
        second_gff = None

    for chunk in gffs:

        if args.read_twice:
            second_gff_chunk = next(second_gff)
        else:
            chunk = list(chunk)
            second_gff_chunk = chunk

        if args.method == "hierarchy":

            convert_hierarchy(chunk, second_gff_chunk, args)
        elif args.method == "set-field":
            gene_id_pattern = "%%(%s)s" % args.gene_field_or_pattern
            transcript_id_pattern = "%%(%s)s" % args.transcript_field_or_pattern
            convert_set(chunk, gene_id_pattern, transcript_id_pattern, args)
        elif args.method == "set-pattern":
            convert_set(chunk, args.gene_field_or_pattern,
                        args.transcript_field_or_pattern, args)
        elif args.method == "set-none":
            convert_set(chunk, None, None, args)

    # write footer and output benchmark information.
    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""Convert fastq files to fasta files

Implemeted methods
==================

ont2pacbio
----------

Convert long-read fastq files, for example from ONT data,
to pacbio formatted fasta files that can be used with the
dazzler toolkit. Pacbio fasta files have the following
naming convention for the fastq header line::

    ><name>/<well>/<start>_<end> RQ=0.<qv>

where well, start, end and qv are integer values.

The quality value will be set to the average of the base qualities.

"""

import sys
import pysam
import math

import cgatcore.experiment as E


def main(argv=sys.argv):

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument(
        "-i", "--input-fastq", dest="input_fastq_file", type=str,
        help="input fastq file")

    parser.add_argument(
        "-m", "--method", dest="method", type=str,
        choices=["ont2pacbio"],
        help="methods to apply ")

    parser.set_defaults(
        input_fastq_file=None,
        line_width=80,
        method=None,
    )

    (args, unknown) = E.start(parser,
                              argv,
                              add_output_options=True,
                              unknowns=True)

    if len(unknown) == 1:
        args.input_fastq_file = unknown[0]

    if args.input_fastq_file == "-":
        args.input_fastq_file = args.stdin

    outf = args.stdout
    line_width = args.line_width
    well_no = 0
    for record in pysam.FastqFile(args.input_fastq_file):
        well_no += 1
        quals = record.get_quality_array()
        seq = record.sequence
        qv = int(math.floor(sum(quals) / len(quals)))
        outf.write(">{}/{}/{}_{} RQ=0.{}\n".format(
            "test", well_no, 1, len(seq) + 1, qv))
        for x in range(0, len(seq), line_width):
            outf.write(seq[x:x + line_width] + "\n")

    E.stop()

if __name__ == "__main__":
    sys.exit(main())
